#!/usr/bin/env python
"""
Module containing the base class for Interference Alignment (IA)
Algorithms.

This module should probably only be imported in the other modules inside
the 'ia' package that implement the IA algorithms.
"""

from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional, Sequence, TypeVar, Union, cast

import numpy as np

import pyphysim.channels.multiuser as muchannels

from ..util.conversion import linear2dB
from ..util.misc import leig, randn_c_RS

NumberOrArray = TypeVar("NumberOrArray", np.ndarray, float)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Base Class for all IA Algorithms xxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class IASolverBaseClass:  # pylint: disable=R0902
    """
    Base class for all Interference Alignment Algorithms.

    At least the `_updateW`, `_updateF` and `solve` methods must be
    implemented in the subclasses of IASolverBaseClass, where the `solve`
    method uses the `_updateW` and `_updateF` methods in its
    implementation.

    The implementation of the `_updateW` method should call the
    `_clear_receive_filter` method in the beginning and after that set
    either the _W or the _W_H variables with the correct value.

    The implementation of the `_updateF` method should call the
    clear_precoder_filter in the beginning and after that set _W variable
    with the correct precoder (normalized to have a Frobenius norm equal to
    one.

    The implementation of the `_updateF` method must set the _F variable
    with the correct value.

    Another method that can be implemented is the get_cost method. It
    should return the cost of the current IA solution. What is
    considered "the cost" varies from one IA algorithm to another,
    but should always be a real non-negative number. If get_cost is not
    implemented a value of -1 is returned.

    Parameters
    ----------
    multiUserChannel : muchannels.MultiUserChannelMatrix
        The multiuser channel.
    """

    # The IASolverBaseClass is an abstract class and the 'solve' method
    # (marked as abstract) must be implemented in a subclass.
    __metaclass__ = ABCMeta

    def __init__(self, multiUserChannel: muchannels.MultiUserChannelMatrix):
        # xxxxxxxxxx Private attributes xxxxxxxxxxxxxxx
        if not isinstance(multiUserChannel, muchannels.MultiUserChannelMatrix):
            raise ValueError("multiUserChannel must be an object of the "
                             "MultiUserChannelMatrix class (or a subclass).")
        # Channel of all users
        self._multiUserChannel = multiUserChannel

        # Number of streams per user
        self._Ns: Optional[np.ndarray] = None
        # Power of each user (P is an 1D numpy array). If not set (_P is
        # None), then a power of 1 will be used for each transmitter.
        self._P: Optional[np.ndarray] = None

        # xxxxxxxxxx Precoder and receive filters xxxxxxxxxxxxxxxxxxxxxxxxx
        # These are numpy arrays of numpy arrays

        # Precoder: One precoder for each user
        self._F: Optional[np.ndarray] = None
        # Precoder: Same as _F, but scaled with the correct power value in
        # self.P
        self._full_F: Optional[np.ndarray] = None
        # Receive filter: One for each user
        self._W: Optional[np.ndarray] = None
        # Receive filter (hermitian): One for each user
        self._W_H: Optional[np.ndarray] = None
        self._full_W_H: Optional[np.ndarray] = None
        self._full_W: Optional[np.ndarray] = None
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxx Other member variables xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # RandomState object used to randomize the precoder
        self._rs = np.random.RandomState()
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def _clear_receive_filter(self) -> None:
        """
        Clear the receive filter.

        This should be called in the beginning of the implementation of the
        `_updateW` method in subclasses.
        """
        self._W = None
        self._W_H = None
        self._full_W_H = None
        self._full_W = None

    def _clear_precoder_filter(self) -> None:
        """
        Clear the precoder filter.

        This should be called in the beginning of the implementation of the
        `_updateF` method in subclasses.
        """
        self._F = None
        self._full_F = None

    def clear(self) -> None:
        """
        Clear the IA Solver object.

        All member attributes that are updated during the solve method,
        such as the precoder and receive filters, will be cleared. The
        other attributes that correspond to "configuration" such as the
        channel object won't be changed.

        Notes
        -----
        You should overwrite this method in subclasses that pass parameters
        to the __init__ method, since here we call __init__ without
        arguments which is probably not what you want.
        """
        # The F and W variables will be numpy arrays of numpy arrays.
        self._clear_precoder_filter()  # Set _F and _full_F to None
        self._clear_receive_filter()  # Set _W, _W_H and _full_W_H to None
        self._P = None
        self._Ns = None

        # Don't clear the self._noise_var attribute

    def get_cost(self) -> float:  # pylint: disable=R0201
        """
        Get the current cost of the IA Solution.

        This method should be implemented in subclasses and return a number
        greater than or equal to zero..

        Returns
        -------
        cost : float
            The Cost of the current IA solution (a real non-negative
            number).
        """
        return -1

    @property
    def noise_var(self) -> float:
        """
        Get method for the noise_var property.

        Returns
        -------
        float
            The noise variance (a real non-negative number).
        """
        noise_var = self._multiUserChannel.noise_var
        if noise_var is None:
            return 0.0

        return noise_var

    @property
    def F(self) -> np.ndarray:
        """
        Transmit precoder of all users.

        Returns
        -------
        np.ndarray
            The precoders of all users (a 1D numpy array of 2D numpy
            arrays).
        """
        return self._F

    @property
    def full_F(self) -> np.ndarray:
        """
        Transmit precoder of all users.

        Returns
        -------
        np.ndarray
            The precoders of all users (a 1D numpy array of 2D numpy
            arrays).
        """
        if self._full_F is None:
            self._full_F = self._F * np.sqrt(self.P)
        return self._full_F

    # noinspection PyUnresolvedReferences
    def set_precoders(self,
                      F: Optional[Sequence[np.ndarray]] = None,
                      full_F: Optional[Sequence[np.ndarray]] = None,
                      P: Optional[np.ndarray] = None) -> None:
        """
        Set the precoders of each user.

        Either `F` or `full_F` (or both of them) must be provided.

        If only `full_F` is provided then the value of `F` will be
        calculated from `full_F`.

        In any case, the value of `self.Ns` will be updated according to
        the dimensions of the `F` or `full_F`.

        Parameters
        ----------
        F : np.ndarray | list[np.ndarray], optional
            A numpy array where each element is the (normalized) precoder
            (a 2D numpy array) of one user.
        full_F : np.ndarray | list[np.ndarray], optional
            A numpy array where each element is the precoder (a 2D numpy
            array) of one user.
        P : np.ndarray, optional
            The maximum transmit power. If not provided the current value
            of self.P will be kept as it is.
        """
        if F is None and full_F is None:
            raise RuntimeError("Either 'F' or 'full_F' must be provided.")

        self._clear_precoder_filter()

        if P is not None:
            self._P = P

        self._full_F = full_F

        if F is None:
            assert (full_F is not None)
            K = len(full_F)
            self._F = np.empty(K, dtype=np.ndarray)
            for k in range(K):
                self._F[k] = full_F[k] / np.linalg.norm(full_F[k], 'fro')
        else:
            self._F = F

        # Update the number of streams
        self._Ns = np.empty(self.K, dtype=int)
        for k in range(self.K):
            self._Ns[k] = self._F[k].shape[1]

    # The W property should return a receive filter that can be directly
    # multiplied (no need to calculate the hermitian of W) by the received
    # signal to cancel interference and compensate the effect of the
    # channel.
    @property
    def W(self) -> np.ndarray:
        """
        Receive filter of all users.

        Returns
        -------
        np.ndarray
            The receive filter of all users. (a 1D numpy array of 2D numpy
            arrays).
        """
        # If self._W is None but self._W_H is not None than we need to
        # update self_W from self._W_H.
        if self._W is None:
            if self._W_H is not None:
                self._W = np.empty(self.K, dtype=np.ndarray)
                for k in range(self.K):
                    self._W[k] = self._W_H[k].conj().T
        return self._W

    @property
    def W_H(self) -> np.ndarray:
        """
        Get method for the W_H property.

        Returns
        -------
        np.ndarray
            The conjugate of the receive filter of all users. (a 1D numpy
            array of 2D numpy arrays).
        """
        # If self._W_H is None but self._W is not None than we need to
        # update self_W_H from self._W.
        if self._W_H is None:
            if self._W is not None:
                self._W_H = np.empty(self.K, dtype=np.ndarray)
                for k in range(self.K):
                    self._W_H[k] = self._W[k].conj().T
        return self._W_H

    @property
    def full_W_H(self) -> np.ndarray:
        """
        Get method for the full_W_H property.

        The full_W_H property returns the equivalent filter of the IA
        filter plus the post processing filter.

        Returns
        -------
        np.ndarray
            The equivalent filter of the IA filter plus the post processing
            filter.
        """
        if self._full_W_H is None:
            if self.W_H is not None:
                self._full_W_H = np.empty(self.K, dtype=np.ndarray)
                for k in range(self.K):
                    # Equivalent channel with the effect of the precoder,
                    # channel and receive filter
                    Hieq = self._calc_equivalent_channel(k)
                    # TODO: Put this in a try-except block for the case
                    # that Hieq is singular (a linalg.LinAlgError exception
                    # is thrown). In order to handle the exception, you
                    # could set full_W_H to just W_H or you could try the
                    # stream reduction (but stream reduction should be
                    # performed at the precoders too)
                    self._full_W_H[k] = np.linalg.solve(Hieq, self.W_H[k])

        return self._full_W_H

    @property
    def full_W(self) -> np.ndarray:
        """
        Get method for the full_W property.

        The full_W property returns the equivalent filter of the IA
        filter plus the post processing filter.

        Returns
        -------
        np.ndarray
            the equivalent filter of the IA filter plus the post processing
            filter.
        """
        if self._full_W is None:
            self._full_W = np.empty(self.K, dtype=np.ndarray)
            for k in range(self.K):
                self._full_W[k] = self.full_W_H[k].conj().T
        return self._full_W

    def set_receive_filters(self,
                            W_H: Optional[Sequence[np.ndarray]] = None,
                            W: Optional[Sequence[np.ndarray]] = None) -> None:
        """
        Set the receive filters.

        You only need to pass either `W_H` or `W`, but not both of them,
        since one is calculated from the other.

        Parameters
        ----------
        W_H : np.ndarray | list[np.ndarray]
            A numpy array where each element is the receive filter (a 2D
            numpy array) of one user. This is a 1D numpy array of 2D numpy
            arrays.
        W : np.ndarray | list[np.ndarray]
            A numpy array where each element is the receive filter (a 2D
            numpy array) of one user. This is a 1D numpy array of 2D numpy
            arrays.
        """
        self._clear_receive_filter()

        if W is None and W_H is None:
            raise RuntimeError("Either 'W' or 'W_H' must be provided.")

        if W is not None and W_H is not None:
            raise RuntimeError("Either 'W' or 'W_H' must be provided ("
                               "but not both of them.")

        self._W = W
        self._W_H = W_H

    def _calc_equivalent_channel(self, k: int) -> np.ndarray:
        """
        Calculates the equivalent channel for user :math:`k` considering
        the effect of the precoder (including transmit power),
        the actual channel, and the receive filter (without power
        compensation).

        Parameters
        ----------
        k : int
            The index of the desired user.

        Returns
        -------
        Hk_eq : np.ndarray
            The equivalent channel.

        Notes
        -----
        This method is used only internally in order to calculate the "W"
        get property so that the returned filter W compensates the effect
        of the direct channel.
        """
        # Note that here Wk_H is the self.Wk_H property and not the
        # self.full_W_H property. Since _calc_equivalent_channel is used in
        # the full_W_H get property if we had used self.full_W_H here we
        # would get an infinity recursion.
        Wk_H = self.W_H[k]
        full_Fk = self.full_F[k]
        Hkk = self._get_channel(k, k)
        Hk_eq = Wk_H.dot(Hkk.dot(full_Fk))
        return Hk_eq

    @property
    def P(self) -> np.ndarray:
        """
        Transmit power of all users.

        Returns
        -------
        np.ndarray
            The power of all users.
        """
        if self._P is None:
            return np.ones(self.K, dtype=float)

        return self._P

    @P.setter
    def P(self, value: Optional[NumberOrArray]) -> None:
        """Transmit power of all users.

        Parameters
        ----------
        value : float | np.ndarray
            The new power of all users.
        """
        if value is None:
            # Note that if self._P is None then the getter property will
            # return a numpy array of ones with the appropriated size.
            self._P = None
        elif np.isscalar(value):
            if value > 0.0:
                self._P = np.ones(self.K, dtype=float) * value
            else:
                raise ValueError("P cannot be negative or equal to zero.")
        else:
            assert (not isinstance(value, float))
            if len(value) != self.K:
                raise ValueError("P must be set to a sequence of length K")

            value = np.array(value)
            assert (isinstance(value, np.ndarray))
            if np.all(value > 0.0):
                self._P = np.array(value)
            else:
                raise ValueError("P cannot be negative or equal to zero.")

    @property
    def Ns(self) -> np.ndarray:
        """
        Number of streams of all users.

        Returns
        -------
        Ns : np.ndarray
            Number of streams of all users.
        """
        return self._Ns

    # xxxxx Properties to read the channel related variables xxxxxxxxxxxxxx
    @property
    def K(self) -> int:
        """
        The number of users.

        Returns
        -------
        K : int
            The number of users.
        """
        K = self._multiUserChannel.K
        assert K > 0, ("You must initialize the channel object before"
                       " using the IA solver object.")
        return K

    @property
    def Nr(self) -> np.ndarray:
        """
        Number of receive antennas of all users.

        Returns
        -------
        Nr : np.ndarray
            Number of receive antennas of all users.
        """
        return self._multiUserChannel.Nr

    @property
    def Nt(self) -> np.ndarray:
        """
        Number of transmit antennas of all users.

        Returns
        -------
        Nt : np.ndarray
            Number of transmit antennas of all users.
        """
        return self._multiUserChannel.Nt

    def randomizeF(self,
                   Ns: Union[int, List[int], Sequence[int]],
                   P: Optional[np.ndarray] = None) -> None:
        """
        Generates a random precoder for each user.

        Parameters
        ----------
        Ns : int | list[int] | np.ndarray
            Number of streams of each user.
        P : np.ndarray, optional
            Power of each user. If not provided, a value of 1 will be used
            for each user.
        """
        self._clear_precoder_filter()

        if isinstance(Ns, int):
            Ns = np.ones(self.K, dtype=int) * Ns
        assert (not isinstance(Ns, int))

        self.P = P

        # Local function that returns a normalized version of the input
        # numpy array
        def normalized(A: np.ndarray) -> np.ndarray:
            return A / np.linalg.norm(A, 'fro')

        self._F = np.zeros(self.K, dtype=np.ndarray)
        for k in range(self.K):
            self._F[k] = normalized(randn_c_RS(self._rs, self.Nt[k], Ns[k]))

        # This will create a new array so that we can modify self._Ns
        # internally without changing the original Ns variable passed to
        # the randomizeF method.
        self._Ns = np.array(Ns)

    # This method is just an alias for the get_channel method of the
    # multiuserchannel object associated with the IA Solver.xs
    def _get_channel(self, k: int, l: int) -> np.ndarray:
        """
        Get the channel from transmitter l to receiver k.

        Parameters
        ----------
        l : int
            Transmitting user.
        k : int
            Receiving user.

        Returns
        -------
        H : np.ndarray
            The channel matrix between transmitter l and receiver k.
        """
        return self._multiUserChannel.get_Hkl(k, l)

    def _get_channel_rev(self, k: int, l: int) -> np.ndarray:
        """
        Get the channel from transmitter l to receiver k in the reverse
        network.

        Let the matrix :math:`\\mtH_{kl}` be the channel matrix between the
        transmitter :math:`l` to receiver :math:`k` in the direct
        network. The channel matrix between the transmitter :math:`l` to
        receiver :math:`k` in the reverse network, denoted as
        :math:`\\overleftarrow{\\mtH}_{kl}`, is then given by
        :math:`\\overleftarrow{\\mtH}_{kl} = \\mtH_{lk}^\\dagger` where
        :math:`\\mtA^\\dagger` is the conjugate transpose of :math:`\\mtA`.

        Parameters
        ----------
        l : int
            Transmitting user of the reverse network.
        k : int
            Receiving user of the reverse network.

        Returns
        -------
        H : np.ndarray
            The channel matrix between transmitter l and receiver k in the
            reverse network.

        Notes
        -----
        See Section III of [Cadambe2008]_ for details.
        """
        return self._get_channel(l, k).transpose().conjugate()

    # noinspection PyPep8
    def calc_Q(self, k: int) -> np.ndarray:
        """
        Calculates the interference covariance matrix at the :math:`k`-th
        receiver.

        The interference covariance matrix at the :math:`k`-th receiver,
        :math:`\\mtQ k`, is given by

            .. math::

               \\mtQ k = \\sum_{j=1, j \\neq k}^{K} \\frac{P_j}{Ns_j} \\mtH_{kj} \\mtF_j \\mtF_j^H \\mtH_{kj}^H

        where :math:`P_j` is the transmit power of transmitter :math:`j`,
        and :math:`Ns_j` is the number of streams for user :math:`j`.

        Parameters
        ----------
        k : int
            Index of the desired receiver.

        Returns
        -------
        Qk : np.ndarray
            The interference covariance matrix at receiver :math:`k`.

        Notes
        -----
        This is impacted by the self.P attribute.
        """
        Qk = self._multiUserChannel.calc_Q(k, self.full_F)
        return Qk

    # This method must be tested in a subclass of IASolverBaseClass, since
    # we need the receive filter and IASolverBaseClass does not know how to
    # calculate it
    def calc_Q_rev(self, k: int) -> np.ndarray:
        """
        Calculates the interference covariance matrix at the :math:`k`-th
        receiver in the reverse network.

        Parameters
        ----------
        k : int
            Index of the desired receiver.

        Returns
        -------
        Qk_rev : np.ndarray
            The interference covariance matrix at receiver :math:`k` in the
            reverse network.

        See also
        --------
        calc_Q
        """
        P = self.P
        interfering_users = set(range(self.K)) - {k}
        Qk = np.zeros([self.Nt[k], self.Nt[k]], dtype=complex)

        for l in interfering_users:
            # The lets make sure the receive filter norm is equal to one so
            # that we can correctly scale it to the desired power.
            assert (isinstance(self._W, np.ndarray))
            assert np.linalg.norm(self._W[l], 'fro') - 1.0 < 1e-6
            Hkl_F_rev = np.dot(self._get_channel_rev(k, l), self._W[l])
            Qk = Qk + np.dot(P[l] * Hkl_F_rev, Hkl_F_rev.conjugate().T)

        return Qk

    # noinspection PyPep8
    def calc_remaining_interference_percentage(self,
                                               k: int,
                                               Qk: Optional[np.ndarray] = None
                                               ) -> float:
        """
        Calculates the percentage of the interference in the desired signal
        space according to equation (30) in [Cadambe2008]_.

        The percentage :math:`p_k` of the interference in the desired
        signal space is given by

            .. math::

               p_k = \\frac{\\sum_{j=1}^{Ns[k]} \\lambda_j [\\mtQ k]}{Tr[\\mtQ k]}

        where :math:`\\lambda_j[\\mtA]` denotes the :math:`j`-th smallest
        eigenvalue of :math:`\\mtA`.

        Parameters
        ----------
        k : int
            The index of the desired user.
        Qk : np.ndarray
            The covariance matrix of the remaining interference at receiver
            k. If not provided, it will be automatically calculated. In
            that case, the `P` attribute will also be taken into account if
            it is set.

        Returns
        -------
        float
            The percentage of the interference in the desired signal space.

        Notes
        -----
        `Qk` must be a symmetric matrix so that its eigenvalues are real and
        positive (any covariance matrix is a symmetric matrix).
        """
        # $$p_k = \frac{\sum_{j=1}^{Ns[k]} \lambda_j [\mtQ k]}{Tr[\mtQ k]}$$

        if Qk is None:
            Qk = self.calc_Q(k)

        [_, D] = leig(Qk, self.Ns[k])
        pk = np.sum(np.abs(D)) / np.trace(np.abs(Qk))
        return cast(float, pk)

    def calc_SINR_old(self) -> np.ndarray:
        """
        Calculates the SINR values (in linear scale) of all streams of all
        users with the current IA solution.

        The noise variance used will be the value of the noise_var
        property, which, if not explicitly set, will use the
        noise_var property of the multiuserchannel object.

        This method is deprecated since it's not the correct way to
        calculate the SINR. Use the calc_SINR method instead.

        Returns
        -------
        SINRs : np.ndarray
            The SINR (in linear scale) of all streams of all users. This
            is a 1D numpy array of 1D numpy arrays (of floats).
        """
        K = self.K
        SINRs: np.array = np.empty(K, dtype=np.ndarray)

        for j in range(K):
            numerator = 0.0
            denominator = 0.0
            Wj_H = self.W_H[j]
            for i in range(K):
                Hji = self._get_channel(j, i)
                Fi = self.F[i]
                aux: np.ndarray = np.dot(Wj_H, np.dot(Hji, Fi))
                if i == j:
                    aux = np.dot(aux, aux.transpose().conjugate())
                    # Numerator will be a 1D numpy array with length equal
                    # to the number of streams
                    numerator = numerator + np.diag(np.abs(aux))
                else:
                    denominator = denominator + aux

            # pylint: disable=E1103
            # noinspection PyUnresolvedReferences
            denominator = np.dot(
                denominator,
                denominator.transpose().conjugate())  # type: ignore
            noise_power = self.noise_var * np.dot(Wj_H,
                                                  Wj_H.transpose().conjugate())
            denominator += noise_power
            denominator = np.diag(np.abs(denominator))

            SINRs[j] = numerator / denominator

        return SINRs

    def calc_SINR(self) -> np.ndarray:
        """
        Calculates the SINR values (in linear scale) of all streams of all
        users with the current IA solution.

        The noise variance used will be the value of the noise_var
        property, which, if not explicitly set, will use the
        noise_var property of the multiuserchannel object.

        Returns
        -------
        SINRs : np.ndarray
            The SINR (in linear scale) of all streams of all users. This is a
            1D numpy array of 1D numpy arrays (of floats).
        """
        K = self.K
        SINRs = np.empty(K, dtype=np.ndarray)

        for k in range(self.K):
            Bkl_all_l = self._calc_Bkl_cov_matrix_all_l(k, self.noise_var)
            SINRs[k] = self._calc_SINR_k(k, Bkl_all_l)
        return SINRs

    def calc_SINR_in_dB(self) -> np.ndarray:
        """
        Calculates the SINR values (in dB scale) of all streams of all
        users with the current IA solution.

        The noise variance used will be the value of the noise_var
        property, which, if not explicitly set, will use the
        noise_var property of the multiuserchannel object.

        Returns
        -------
        SINRs : np.ndarray
            The SINR (in dB scale) of all streams of all users. This is a 1D
            numpy array of 1D numpy arrays (of floats).
        """
        K = self.K
        SINRs = np.empty(K, dtype=np.ndarray)

        for k in range(self.K):
            Bkl_all_l = self._calc_Bkl_cov_matrix_all_l(k, self.noise_var)
            SINRs[k] = linear2dB(self._calc_SINR_k(k, Bkl_all_l))
        return SINRs

    def calc_sum_capacity(self) -> float:
        """
        Calculates the sum capacity of the current solution.

        The SINRs are estimated and applied to the Shannon capacity formula

        Returns
        -------
        float
            The sum capacity of the current solution.
        """
        return cast(float, np.sum(np.log2(1 + np.hstack(self.calc_SINR()))))

    # noinspection PyPep8
    def _calc_Bkl_cov_matrix_first_part(self, k: int) -> np.ndarray:
        """
        Calculates the first part in the equation of the Blk covariance matrix
        in equation (28) of [Cadambe2008]_.

        The first part is given by

            .. math::

               \\sum_{j=1}^{K} \\frac{P^{[j]}}{d^{[j]}} \\sum_{d=1}^{d^{[j]}} \\mtH^{[kj]}\\mtV_{\\star d}^{[j]} \\mtV_{\\star d}^{[j]\\dagger} \\mtH^{[kj]\\dagger}

        Note that it only depends on the value of :math:`k`.

        Parameters
        ----------
        k : int
            Index of the desired user.

        Returns
        -------
        Bkl_first_part : np.ndarray
            First part in equation (28) of [Cadambe2008]_.
        """
        # $$\sum_{j=1}^{K} \frac{P^{[j]}}{d^{[j]}} \sum_{d=1}^{d^{[j]}} \mtH^{[kj]}\mtV_{\star d}^{[j]} \mtV_{\star d}^{[j]\dagger} \mtH^{[kj]\dagger}$$
        first_part = 0.0
        for j in range(self.K):
            Hkj = self._get_channel(k, j)
            Vj = self.full_F[j]

            aux = np.dot(Hkj, Vj)

            first_part = first_part + np.dot(aux, aux.conjugate().T)

        return first_part

    # noinspection PyPep8
    def _calc_Bkl_cov_matrix_second_part(self, k: int, l: int) -> np.ndarray:
        """
        Calculates the second part in the equation of the Blk covariance matrix
        in equation (28) of [Cadambe2008]_ (note that it does not include
        the identity matrix).

        The second part is given by

            .. math::

               \\frac{P^{[k]}}{d^{[k]}} \\mtH^{[kk]} \\mtV_{\\star l}^{[k]} \\mtV_{\\star l}^{[k]\\dagger} \\mtH^{[kk]\\dagger}

        Parameters
        ----------
        k : int
            Index of the desired user.
        l : int
            Index of the desired stream.

        Returns
        -------
        second_part : np.ndarray
            Second part in equation (28) of [Cadambe2008]_.
        """
        # $$\frac{P^{[k]}}{d^{[k]}} \mtH^{[kk]} \mtV_{\star l}^{[k]} \mtV_{\star l}^{[k]\dagger} \mtH^{[kk]\dagger}$$
        Hkk = self._get_channel(k, k)
        Vkl = self.full_F[k][:, l:l + 1]
        aux = np.dot(Hkk, Vkl)
        second_part = np.dot(aux, aux.conjugate().T)

        return second_part

    # noinspection PyPep8
    def _calc_Bkl_cov_matrix_all_l(self,
                                   k: int,
                                   noise_power: Optional[float] = None
                                   ) -> np.ndarray:
        """
        Calculates the interference-plus-noise covariance matrix for all
        streams at receiver :math:`k` according to equation (28) in
        [Cadambe2008]_.

        The interference-plus-noise covariance matrix for stream :math:`l`
        of user :math:`k` is given by Equation (28) in [Cadambe2008]_,
        which is reproduced below

            .. math::

               \\mtB^{[kl]} = \\sum_{j=1}^{K} \\frac{P^{[j]}}{d^{[j]}} \\sum_{d=1}^{d^{[j]}} \\mtH^{[kj]}\\mtV_{\\star l}^{[j]} \\mtV_{\\star l}^{[j]\\dagger} \\mtH^{[kj]\\dagger} - \\frac{P^{[k]}}{d^{[k]}} \\mtH^{[kk]} \\mtV_{\\star l}^{[k]} \\mtV_{\\star l}^{[k]\\dagger} \\mtH^{[kk]\\dagger} + \\mtI_{N^{[k]}}

        where :math:`P^{[k]}` is the transmit power of transmitter
        :math:`k`, :math:`d^{[k]}` is the number of degrees of freedom of
        user :math:`k`, :math:`\\mtH^{[kj]}` is the channel between
        transmitter :math:`j` and receiver :math:`k`, :math:`\\mtV_{\\star
        l}` is the :math:`l`-th column of the precoder of user :math:`k`
        and :math:`\\mtI_{N^{k}}` is an identity matrix with size equal to
        the number of receive antennas of receiver :math:`k`.

        Parameters
        ----------
        k : int
            Index of the desired user.
        noise_power : float
            Noise power (variance).

        Returns
        -------
        Bkl : np.ndarray
            Covariance matrix of all streams of user k. Each element of the
            returned 1D numpy array is a 2D numpy complex array
            corresponding to the covariance matrix of one stream of user k.

        Notes
        -----

        To be simple, a function that returns the covariance matrix of only
        a single stream "l" of the desired user "k" could be implemented,
        but in the order to calculate the max SINR algorithm we need the
        covariance matrix of all streams and returning them in single
        function as is done here allows us to calculate the first part in
        equation (28) of [Cadambe2008]_ only once, since it is the same for
        all streams.
        """
        # $$\mtB^{[kl]} = \sum_{j=1}^{K} \frac{P^{[j]}}{d^{[j]}} \sum_{d=1}^{d^{[j]}} \mtH^{[kj]}\mtV_{\star l}^{[j]} \mtV_{\star l}^{[j]\dagger} \mtH^{[kj]\dagger} - \frac{P^{[k]}}{d^{[k]}} \mtH^{[kk]} \mtV_{\star l}^{[k]} \mtV_{\star l}^{[k]\dagger} \mtH^{[kk]\dagger} + \sigma_n^2 \mtI_{N^{[k]}}$$
        if noise_power is None:
            noise_power = self.noise_var

        assert (self._Ns is not None)
        Bkl_all_l = np.empty(self._Ns[k], dtype=np.ndarray)
        first_part = self._calc_Bkl_cov_matrix_first_part(k)
        for l in range(self._Ns[k]):
            second_part = self._calc_Bkl_cov_matrix_second_part(k, l)
            Bkl_all_l[l] = first_part - second_part + (noise_power *
                                                       np.eye(self.Nr[k]))

        return Bkl_all_l

    def _calc_SINR_k(self, k: int,
                     Bkl_all_l: Sequence[np.ndarray]) -> np.ndarray:
        """
        Calculates the SINR of all streams of user 'k'.

        Parameters
        ----------
        k : int
            Index of the desired user.
        Bkl_all_l : list[np.ndarray] | nd.ndarray
            A sequence (1D numpy array, a list, etc) of 2D numpy arrays
            corresponding to the Bkl matrices for all 'l's.

        Returns
        -------
        SINR_k : np.ndarray
            The SINR for the different streams of user k.
        """
        Hkk = self._get_channel(k, k)
        Vk = self.full_F[k]
        Uk_H = self.full_W_H[k]

        SINR_k = np.empty(self.Ns[k], dtype=float)

        for l in range(self.Ns[k]):
            Vkl = Vk[:, l:l + 1]
            Ukl_H = Uk_H[l:l + 1, :]
            Ukl = Ukl_H.conj().T
            aux = np.dot(Ukl_H, np.dot(Hkk, Vkl))
            numerator = np.dot(aux, aux.transpose().conjugate())
            denominator = np.dot(Ukl_H, np.dot(Bkl_all_l[l], Ukl))
            SINR_kl = numerator.item() / denominator.item()
            # The imaginary part should be negligible
            SINR_k[l] = np.abs(SINR_kl)

        return SINR_k

    @abstractmethod
    def solve(self,
              Ns: Union[int, np.ndarray],
              P: Optional[np.ndarray] = None) -> Any:  # pragma: no cover
        """
        Find the IA solution.

        This method must be implemented in a subclass and should updates
        the 'F' and 'W' member variables.

        Parameters
        ----------
        Ns : int | np.ndarray
            Number of streams of each user.
        P : np.ndarray
            Power of each user. If not provided, a value of 1 will be used
            for each user.

        Notes
        -----
        This function should be implemented in the derived classes
        """
        raise NotImplementedError("solve: Not implemented")



#!/usr/bin/env python
"""Module containing single user channels. """

import math
from typing import List, Optional, Union

import numpy as np

from . import fading
from .fading_generators import JakesSampleGenerator, RayleighSampleGenerator

# Either a jakes or a Rayleigh model can be used
FadingGenerator = Union[JakesSampleGenerator, RayleighSampleGenerator]

# Type representing something that can be used to index a numpy array
Indexes = Union[np.ndarray, List[int], slice]


class SuChannel:
    """
    Single User channel corresponding to a Tapped Delay Line channel model,
    which corresponds to a multipath channel. You can use a single tap in
    order to get a flat fading channel.

    You can create a new SuChannel object either specifying the channel
    profile or specifying both the channel tap powers and delays. If
    only the fading_generator is specified then a single tap with
    unitary power and delay zero will be assumed, which corresponds to a
    flat fading channel model.

    Parameters
    ----------
    fading_generator : FadingGenerator
        The instance of a fading generator in the `fading_generators`
        module.  It should be a subclass of FadingSampleGenerator. The
        fading generator will be used to generate the channel samples. If
        not provided then RayleighSampleGenerator will be used
    channel_profile : fading.TdlChannelProfile
        The channel profile, which specifies the tap powers and delays.
    tap_powers_dB : np.ndarray
        The powers of each tap (in dB). Dimension: `L x 1`
        Note: The power of each tap will be a negative number (in dB).
    tap_delays : np.ndarray
        The delay of each tap (in seconds). Dimension: `L x 1`
    Ts : float, optional
        The sampling interval.
    """
    def __init__(self,
                 fading_generator: Optional[FadingGenerator] = None,
                 channel_profile: Optional[fading.TdlChannelProfile] = None,
                 tap_powers_dB: Optional[np.ndarray] = None,
                 tap_delays: Optional[np.ndarray] = None,
                 Ts: Optional[float] = None):

        fading_generator_param: FadingGenerator
        if fading_generator is None:
            fading_generator_param = RayleighSampleGenerator()
            if channel_profile is None and Ts is None:
                Ts = 1.0
        else:
            fading_generator_param = fading_generator

        if (channel_profile is None and tap_powers_dB is None
                and tap_delays is None):
            # Only the fading generator was provided. Let's assume a flat
            # fading channel
            self._tdlchannel = fading.TdlChannel(fading_generator_param,
                                                 tap_powers_dB=np.zeros(1),
                                                 tap_delays=np.zeros(1),
                                                 Ts=Ts)
        else:
            # More parameters were provided. We will have then a TDL
            # channel model. Let's just pass these parameters to the
            # base class.
            self._tdlchannel = fading.TdlChannel(fading_generator_param,
                                                 channel_profile,
                                                 tap_powers_dB, tap_delays, Ts)

        # Path loss which will be multiplied by the impulse response when
        # corrupt_data is called
        self._pathloss_value: Optional[float] = None

    def set_pathloss(self, pathloss_value: Optional[float] = None) -> None:
        """
        Set the path loss (IN LINEAR SCALE).

        The path loss will be accounted when calling the corrupt_data
        method.

        If you want to disable the path loss, set `pathloss_value` to
        None.

        Parameters
        ----------
        pathloss_value : float | None
            The path loss (IN LINEAR SCALE) from the transmitter to the
            receiver. If you want to disable the path loss then set it to
            None.

        Notes
        -----
        Note that path loss is a power relation, which means that the
        channel coefficients will be multiplied by the square root of
        elements in `pathloss_value`.
        """
        if pathloss_value is not None:
            if pathloss_value < 0 or pathloss_value > 1:
                raise ValueError("Pathloss must be between 0 and 1")

        self._pathloss_value = pathloss_value

    def set_num_antennas(self, num_rx_antennas: int,
                         num_tx_antennas: int) -> None:
        """
        Set the number of transmit and receive antennas for MIMO
        transmission.

        Set both `num_rx_antennas` and `num_tx_antennas` to None for SISO
        transmission

        Parameters
        ----------
        num_rx_antennas : int
            The number of receive antennas.
        num_tx_antennas : int
            The number of transmit antennas.
        """
        self._tdlchannel.set_num_antennas(num_rx_antennas, num_tx_antennas)

    def corrupt_data(self, signal: np.ndarray) -> np.ndarray:
        """
        Transmit the signal through the TDL channel.

        Parameters
        ----------
        signal : np.ndarray
            The signal to be transmitted.

        Returns
        -------
        np.ndarray
            The received signal after transmission through the TDL channel.
        """
        # output = super(SuChannel, self).corrupt_data(signal)
        output = self._tdlchannel.corrupt_data(signal)

        if self._pathloss_value is not None:
            # noinspection PyTypeChecker
            output *= math.sqrt(self._pathloss_value)

        return output

    def corrupt_data_in_freq_domain(
            self,
            signal: np.ndarray,
            fft_size: int,
            carrier_indexes: Optional[Indexes] = None) -> np.ndarray:
        """
        Transmit the signal through the TDL channel, but in the frequency
        domain.

        This is ROUGHLY equivalent to modulating `signal` with OFDM using
        `fft_size` subcarriers, transmitting through a regular TdlChannel,
        and then demodulating with OFDM to recover the received signal.

        One important difference is that here the channel is considered
        constant during the transmission of `fft_size` elements in
        `signal`, and then it is varied by the equivalent of the variation
        for that number of elements. That is, the channel is block static.

        Parameters
        ----------
        signal : np.ndarray
            The signal to be transmitted.
        fft_size : int
            The size of the Fourier transform to get the frequency
            response.
        carrier_indexes : slice | np.ndarray
            The indexes of the subcarriers where signal is to be
            transmitted. If it is None assume all subcarriers will be
            used. This can be a slice object or a numpy array of integers.

        Returns
        -------
        np.ndarray
            The received signal after transmission through the TDL channel
        """
        output = self._tdlchannel.corrupt_data_in_freq_domain(
            signal, fft_size, carrier_indexes)

        if self._pathloss_value is not None:
            # noinspection PyTypeChecker
            output *= math.sqrt(self._pathloss_value)
        return output

    def get_last_impulse_response(self) -> fading.TdlImpulseResponse:
        """
        Get the last generated impulse response.

        A new impulse response is generated when the method `corrupt_data`
        is called. You can use the `get_last_impulse_response` method to
        get the impulse response used to corrupt the last data.

        Returns
        -------
        fading.TdlImpulseResponse
            The impulse response of the channel that was used to corrupt
            the last data.
        """
        if self._pathloss_value is None:
            return self._tdlchannel.get_last_impulse_response()

        return math.sqrt(self._pathloss_value) * \
            self._tdlchannel.get_last_impulse_response()

    @property
    def switched_direction(self) -> bool:
        """
        Get the value of `switched_direction`.

        Returns
        -------
        bool
            True if direction is switched and False otherwise.
        """
        return self._tdlchannel.switched_direction

    @switched_direction.setter
    def switched_direction(self, value: bool) -> None:
        """
        Set the value of `switched_direction`.

        Parameters
        ----------
        value : bool
            True to switch directions of false to use original direction.
        """
        self._tdlchannel.switched_direction = value

    @property
    def num_taps(self) -> int:
        """
        Get the number of taps in the profile.

        Returns
        -------
        int
            The number of taps in the channel (not including any zero
            padding).
        """
        return self._tdlchannel.num_taps

    @property
    def num_taps_with_padding(self) -> int:
        """
        Get the number of taps in the profile including zero-padding
        when the profile is discretized.

        If the profile is not discretized an exception is raised.

        Returns
        -------
        int
            The number of taps in the channel (including any zero padding).
        """
        return self._tdlchannel.num_taps_with_padding

    @property
    def channel_profile(self) -> fading.TdlChannelProfile:
        """
        Return the channel profile.

        Returns
        -------
        fading.TdlChannelProfile
            The channel profile.
        """
        return self._tdlchannel.channel_profile

    @property
    def num_tx_antennas(self) -> int:
        """
        Get the number of transmit antennas.

        Returns
        -------
        int
            The number of transmit antennas.
        """
        return self._tdlchannel.num_tx_antennas

    @property
    def num_rx_antennas(self) -> int:
        """
        Get the number of receive antennas.

        Returns
        -------
        int
            The number of receive antennas.
        """
        return self._tdlchannel.num_rx_antennas


class SuMimoChannel(SuChannel):
    """
    Single User channel corresponding to a Tapped Delay Line channel model,
    which corresponds to a multipath channel. You can use a single tap in
    order to get a flat fading channel.

    You can create a new SuMimoChannel object either specifying the
    channel profile or specifying both the channel tap powers and
    delays. If only the fading_generator is specified then a single tap
    with unitary power and delay zero will be assumed, which corresponds
    to a flat fading channel model.

    Parameters
    ----------
    num_antennas : int
        Number of transmit and receive antennas.
    fading_generator : FadingGenerator
        The instance of a fading generator in the `fading_generators`
        module.  It should be a subclass of FadingSampleGenerator. The
        fading generator will be used to generate the channel samples. If
        not provided then RayleighSampleGenerator will be used
    channel_profile : fading.TdlChannelProfile
        The channel profile, which specifies the tap powers and delays.
    tap_powers_dB : np.ndarray
        The powers of each tap (in dB). Dimension: `L x 1`
        Note: The power of each tap will be a negative number (in dB).
    tap_delays : np.ndarray
        The delay of each tap (in seconds). Dimension: `L x 1`
    Ts : float, optional
        The sampling interval.
    """
    def __init__(self,
                 num_antennas: int,
                 fading_generator: Optional[FadingGenerator] = None,
                 channel_profile: Optional[fading.TdlChannelProfile] = None,
                 tap_powers_dB: Optional[np.ndarray] = None,
                 tap_delays: Optional[np.ndarray] = None,
                 Ts: Optional[float] = None):
        # Before calling supper to initialize the base class we will set
        # the shape of the fading generator
        fading_generator_param: FadingGenerator
        if fading_generator is None:
            fading_generator_param = RayleighSampleGenerator()
            if channel_profile is None and Ts is None:
                Ts = 1.0
        else:
            fading_generator_param = fading_generator

        # Set the shape of the fading generator.
        fading_generator_param.shape = (num_antennas, num_antennas)

        # Initialize attributes from base class
        super(SuMimoChannel,
              self).__init__(fading_generator_param, channel_profile,
                             tap_powers_dB, tap_delays, Ts)



from typing import TypeVar, Any
import numpy as np
import pandas as pd

Array = TypeVar("Array")

X = TypeVar("X", int, float, str, complex, np.ndarray)


def identity(x: X) -> X:
    return x


# if __name__ == '__main__':
y = identity(5)
z = identity(5.5)
x = identity(np.array([1, 2, 3, 4]))

data = data = {"a": [1, 2, 3, 4], "b": [10, 11, 12, 13]}
df = pd.DataFrame(data=data)
a = identity(data)
b = identity(df)
# reveal_type(y)
# reveal_type(z)
# reveal_type(x)
# reveal_type(x.shape)
# reveal_type(a)

# Question in stackoverflow
# tags: [python] [mypy] [typing] [type-hinting]



#!/usr/bin/env python
"""Module implementing different MIMO schemes.

Each MIMO scheme is implemented as a class inheriting from
:class:`MimoBase` and implements at least the methods `encode`, `decode`
and `getNumberOfLayers`.

"""

import math
import warnings
from abc import ABCMeta, abstractmethod
from typing import Optional, cast

import numpy as np

from pyphysim.util.conversion import linear2dB
from pyphysim.util.misc import gmd

__all__ = [
    'MimoBase', 'MisoBase', 'Blast', 'Alamouti', 'MRT', 'MRC', 'SVDMimo',
    'GMDMimo'
]

# TODO: maybe you can use the weave module (inline or blitz methods) from
# scipy to speed up things here.
# See http://docs.scipy.org/doc/scipy/reference/tutorial/weave.html


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Functions to calculate the SINRs xxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def calc_post_processing_SINRs(
        channel: np.ndarray,
        W: np.ndarray,
        G_H: np.ndarray,
        noise_var: Optional[float] = None) -> np.ndarray:
    """
    Calculate the post processing SINRs (in dB) of all streams for a given
    MIMO scheme.

    Parameters
    ----------
    channel : np.ndarray
        The MIMO channel. This should be a 2D numpy array.
    W : np.ndarray
        The precoder for the MIMO scheme. This should be a 2D numpy array.
    G_H : np.ndarray
        The receive filter for the MIMO scheme. This should be a 2D numpy
        array.
    noise_var : float
        The noise variance

    Returns
    -------
    np.ndarray
        The SINR of all streams (in linear scale).
    """
    return linear2dB(
        calc_post_processing_linear_SINRs(channel, W, G_H, noise_var))


def calc_post_processing_linear_SINRs(
        channel: np.ndarray,
        W: np.ndarray,
        G_H: np.ndarray,
        noise_var: Optional[float] = None) -> np.ndarray:
    """
    Calculate the post processing SINRs (in linear scale) of all streams
    for a given MIMO scheme.

    Parameters
    ----------
    channel : np.ndarray
        The MIMO channel. This should be a 2D numpy array.
    W : np.ndarray
        The precoder for the MIMO scheme. This should be a 2D numpy array.
    G_H : np.ndarray
        The receive filter for the MIMO scheme. This should be a 2D
        numpy array.
    noise_var : float
        The noise variance

    Returns
    -------
    np.ndarray
        The SINR of all streams (in linear scale).
    """
    if noise_var is None:  # pragma: nocover
        noise_var = 0.0

    # This matrix will always be square
    channel_eq = np.dot(G_H, channel.dot(W))
    sum_all_antennas = np.sum(channel_eq, axis=1)
    s = np.diag(channel_eq)
    i = sum_all_antennas - s

    S = np.abs(s)**2
    I = np.abs(i)**2

    if isinstance(G_H, np.ndarray):
        # G_H is a numpy array. Lets calculate the norm considering the
        # second axis (axis 1). That is, calculate the norm of each row in
        # G_H. The square of this norm will gives us the amount of noise
        # amplification in each stream.
        N = noise_var * np.linalg.norm(G_H, axis=1)**2
    else:
        # G_H is a single number. The square of its absolute value will
        # gives us the noise amplification of the single stream.
        N = noise_var * abs(G_H)**2

    sinrs = S / (I + N)

    return sinrs


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MimoBase Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MimoBase:
    """
    Base Class for MIMO schemes.

    All subclasses must implement at least the following methods:

    - :meth:`getNumberOfLayers`:
      Should return the number of layers of that specific MIMO scheme
    - :meth:`encode`:
      The encode method must perform everything executed at the transmitter
      for that specific MIMO scheme. This also include the power division
      among the transmit antennas.
    - :meth:`decode`:
      Analogous to the encode method, the decode method must perform
      everything performed at the receiver.

    If possible, subclasses should implement the `_calc_precoder` and
    `_calc_receive_filter` static methods and use them in the
    implementation of `encode` and `decode`. This will allow using the
    `calc_linear_SINRs` and `calc_SINRs` methods to calculate the post
    processing SINRs. Note that calling the `_calcZeroForceFilter` and
    `_calcMMSEFilter` methods in the implementation of the receive filter
    calculation can be useful.

    If you can't implement the `_calc_precoder` and `_calc_receive_filter`
    static methods` (because there is no linear precoder or receive filter
    for the MIMO scheme in the subclass, for instance), then you should
    implement the `calc_linear_SINRs` method in the subclass instead.

    Parameters
    ----------
    channel : np.ndarray | None
        MIMO channel matrix. This should be a 1D or 2D numpy array. The
        allowed dimensions will depend on the particular MIMO scheme
        implemented in a subclass.
    """
    # The MimoBase class is an abstract class and all methods marked as
    # 'abstract' must be implemented in a subclass.
    __metaclass__ = ABCMeta

    def __init__(self, channel: Optional[np.ndarray] = None):
        self._channel = channel

        if channel is not None:
            self.set_channel_matrix(channel)

    def set_channel_matrix(self, channel: np.ndarray) -> None:
        """
        Set the channel matrix.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix. This should be a 1D or 2D numpy array. The
            allowed dimensions will depend on the particular MIMO scheme
            implemented in a subclass.
        """
        self._channel = channel

    @property
    def Nt(self) -> int:
        """
        Get the number of transmit antennas

        Returns
        -------
        int
            The number of transmit antennas.
        """
        assert (self._channel is not None)
        return cast(int, self._channel.shape[1])

    @property
    def Nr(self) -> int:
        """
        Get the number of receive antennas

        Returns
        -------
        int
            The number of receive antennas.
        """
        assert (self._channel is not None)
        return cast(int, self._channel.shape[0])

    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:  # pragma: nocover
        """
        Calculate the linear precoder for the MIMO scheme, if there is any.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.

        Returns
        -------
        W : np.ndarray
            The precoder that can be applied to the input data.
        """
        raise NotImplementedError(
            '_calc_precoder still needs to be implemented')

    @staticmethod
    def _calc_receive_filter(
        channel: np.ndarray,
        noise_var: Optional[float] = None
    ) -> np.ndarray:  # pragma: nocover
        """
        Calculate the receive filter for the MIMO scheme, if there is any.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            The noise variance.

        Returns
        -------
        G_H : np.ndarray
            The receive_filter that can be applied to the input data.
        """
        raise NotImplementedError(
            '_calc_receive_filter still needs to be implemented')

    @abstractmethod
    def getNumberOfLayers(self) -> int:  # pragma: no cover
        """
        Get the number of layers of the MIMO scheme.

        Notes
        -----
        This method must be implemented in each subclass of `MimoBase`.

        Returns
        -------
        int
            The number of layers.
        """
        m = ('getNumberOfLayers still needs to '
             'be implemented in the {0} class')
        raise NotImplementedError(m.format(self.__class__.__name__))

    @staticmethod
    def _calcZeroForceFilter(channel: np.ndarray) -> np.ndarray:
        """
        Calculates the Zero-Force filter to cancel the inter-stream
        interference.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.

        Returns
        -------
        W : np.ndarray
            The Zero-Forcing receive filter.

        Notes
        -----
        The Zero-Force filter basically corresponds to the pseudo-inverse
        of the channel matrix.
        """
        return np.linalg.pinv(channel)

    @staticmethod
    def _calcMMSEFilter(channel: np.ndarray, noise_var: float) -> np.ndarray:
        """
        Calculates the MMSE filter to cancel the inter-stream interference.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            Noise variance.

        Returns
        -------
        W : np.ndarray
            The MMSE receive filter.
        """
        H = channel
        H_H = H.conj().T
        Nt = H.shape[1]
        W = np.linalg.solve(np.dot(H_H, H) + noise_var * np.eye(Nt), H_H)

        return W

    def calc_linear_SINRs(self, noise_var: float) -> np.ndarray:
        """
        Calculate the SINRs (in linear scale) of the multiple streams.

        Parameters
        ----------
        noise_var : float
            The noise variance.

        Returns
        -------
        sinrs : np.ndarray
            The sinrs (in linear scale) of the multiple streams.
        """
        W = self._calc_precoder(self._channel)
        G_H = self._calc_receive_filter(self._channel, noise_var)
        sinrs = calc_post_processing_SINRs(self._channel, W, G_H, noise_var)
        return sinrs

    def calc_SINRs(self, noise_var: float) -> np.ndarray:
        """
        Calculate the SINRs (in dB) of the multiple streams.

        Parameters
        ----------
        noise_var : float
            The noise variance.

        Returns
        -------
        SINRs : np.ndarray
            The SINRs (in dB) of the multiple streams.
        """
        return linear2dB(self.calc_linear_SINRs(noise_var))

    # noinspection PyPep8
    @abstractmethod
    def encode(self, transmit_data: np.ndarray) -> np.ndarray:  # pragma: no cover, pylint: disable=W0613
        """
        Method to encode the transmit data array to be transmitted using
        some MIMO scheme. This method must be implemented in a subclass.

        Parameters
        ----------
        transmit_data : np.ndarray
            The data to be transmitted.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.
        """
        msg = 'encode still needs to be implemented in the {0} class'
        raise NotImplementedError(msg.format(self.__class__.__name__))

    # noinspection PyPep8
    @abstractmethod
    def decode(self, received_data: np.ndarray) -> np.ndarray:  # pragma: no cover, pylint: disable=W0613
        """
        Method to decode the transmit data array to be transmitted using
        some MIMO scheme. This method must be implemented in a subclass.

        Parameters
        ----------
        received_data : np.ndarray
            The received data.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        msg = 'decode still needs to be implemented in the {0} class'
        raise NotImplementedError(msg.format(self.__class__.__name__))


# noinspection PyAbstractClass
class MisoBase(MimoBase):  # pylint: disable=W0223
    """
    Base Class for MISO schemes.

    All subclasses must implement at least the following methods:

    - :meth:`MimoBase.encode`:
      The encode method must perform everything executed at the transmitter
      for that specific MIMO scheme. This also include the power division
      among the transmit antennas.
    - :meth:`MimoBase.decode`:
      Analogous to the encode method, the decode method must perform
      everything performed at the receiver.

    Other optional methods that might be useful implementing in subclasses
    are the `_calc_precoder` and `_calc_receive_filter` methods.

    Parameters
    ----------
    channel : np.ndarray
        MISO channel matrix/vector. MISO schemes are defined for
        scenarios with multiple transmit antennas and a single receive
        antenna. If `channel` is 2D, then the first dimension size must
        be equal to 1.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        MimoBase.__init__(self, channel=None)
        if channel is not None:
            self.set_channel_matrix(channel)

    def set_channel_matrix(self, channel: np.ndarray) -> None:
        """
        Set the channel matrix.

        Parameters
        ----------
        channel : np.ndarray
            MISO channel vector. A MISO scheme is defined for the scenario
            with multiple transmit antennas and a single receive
            antenna. If channel is 2D then the first dimension size must be
            equal to 1.

        Returns
        -------
        None
        """
        # We will store the channel as a 2D numpy to be consistent with the
        # other MIMO classes
        if len(channel.shape) == 1:
            super(MisoBase, self).set_channel_matrix(channel[np.newaxis, :])
        else:
            Nr = channel.shape[0]
            if Nr != 1:
                raise ValueError("The MRT scheme is only defined for the "
                                 "scenario with a single receive antenna")
            # By calling the parent set_channel_matrix method the
            # self._W and self._G_H will be set to None
            super(MisoBase, self).set_channel_matrix(channel)

    def getNumberOfLayers(self) -> int:  # pragma: no cover
        """
        Get the number of layers of the MISO scheme.

        Because a MISO scheme only has one receive antenna then then number
        of layers is always equal to 1.

        Returns
        -------
        int
            The number of layers.
        """
        return 1


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Blast Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class Blast(MimoBase):
    """
    MIMO class for the BLAST scheme.

    The receive filter used will depend on the noise variance (see the
    :meth:`set_noise_var` method). If the noise variance is positive the
    MMSE filter will be used, otherwise noise variance will be ignored and
    the Zero-Forcing filter will be used.

    Parameters
    ----------
    channel : np.ndarray
        MIMO channel matrix.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        """
        Initialized the Blast object.

        If `channel` is not provided you need to call the
        `set_channel_matrix` method before calling the `decode` or the
        `getNumberOfLayers` methods.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        """
        MimoBase.__init__(self, channel)
        self._noise_var: float = 0.0

    def set_channel_matrix(self, channel: np.ndarray) -> None:
        """
        Set the channel matrix.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        """
        Nr, Nt = channel.shape
        if Nt > Nr:
            # Since the number of streams will be equal to the number of
            # transmit antennas, the number of receive antennas should be
            # greater than or equal to the number of transmit antennas. If
            # this is not the case, then you won't be able to recover the
            # streams in the decode method.
            msg = ("The number of transmit antennas for {0} should not be "
                   "greater than the number of receive antennas.").format(
                       self.__class__.__name__)
            warnings.warn(msg)

        super(Blast, self).set_channel_matrix(channel)

    def getNumberOfLayers(self) -> int:
        """
        Get the number of layers of the Blast scheme.

        Returns
        -------
        Nl : int
            Number of layers of the MIMO scheme.
        """
        return self.Nt

    def set_noise_var(self, noise_var: Optional[float]) -> None:
        """
        Set the noise variance for the MMSE receive filter.

        If noise_var is non-positive then the Zero-Force filter will be
        used instead.

        Parameters
        ----------
        noise_var : float | None
            Noise variance for the MMSE filter (if `noise_var` is
            positive). If `noise_var` is 0.0 or None then the Zero-Forcing
            filter will be used.

        Returns
        -------
        None
        """
        if noise_var is None:
            self._noise_var = 0.0

        elif noise_var >= 0.0:
            self._noise_var = noise_var
        else:
            raise ValueError('Noise variance must be a non-negative value.')

    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:
        """
        Calculate the linear precoder for the BLAST scheme.

        The BLAST scheme simple send the data through the multiple streams
        without any particular precoding. Therefore, its linear precoder is
        equivalent to an identity matrix.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.

        Returns
        -------
        W : np.ndarray
            The precoder that can be applied to the input data.
        """
        Nt = channel.shape[1]
        return np.eye(Nt) / math.sqrt(Nt)

    @staticmethod
    def _calc_receive_filter(channel: np.ndarray,
                             noise_var: Optional[float] = None) -> np.ndarray:
        """
        Calculate the receive filter for the MIMO scheme, if there is any.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            The noise variance. If a value is provided then MMSE filter
            will be used. If it is not provided (or None is passes) then
            Zero Force filter will be used.

        Returns
        -------
        G_H : np.ndarray
            The receive_filter that can be applied to the input data.
        """
        Nt = channel.shape[1]

        if noise_var is None:  # pragma: nocover
            noise_var = 0.0

        if noise_var > 0:
            G_H = Blast._calcMMSEFilter(channel, noise_var)
        else:
            G_H = Blast._calcZeroForceFilter(channel)

        return G_H * math.sqrt(Nt)

    def encode(self, transmit_data: np.ndarray) -> np.ndarray:
        """
        Encode the transmit data array to be transmitted using the BLAST
        scheme.

        Parameters
        ----------
        transmit_data : np.ndarray
            A numpy array with a number of elements which is a multiple of
            the number of transmit antennas.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.

        Raises
        ------
        ValueError
            If the number of elements in `transmit_data` is not multiple of
            the number of transmit antennas.
        """
        num_elements = transmit_data.size
        nStreams = self.getNumberOfLayers()
        if num_elements % nStreams != 0:
            # Note this is a single string
            msg = ("Input array number of elements must be a multiple of the"
                   " number of transmit antennas")
            raise ValueError(msg)

        encoded_data = (transmit_data.reshape(
            (nStreams, -1), order='F') / math.sqrt(self.Nt))
        return encoded_data

    def decode(self, received_data: np.ndarray) -> np.ndarray:
        """
        Decode the received data array.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the Blast scheme and
            corrupted by the channel `channel`.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        G_H = self._calc_receive_filter(self._channel, self._noise_var)
        decoded_data = G_H.dot(received_data).reshape(-1, order='F')
        return decoded_data


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MRT xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MRT(MisoBase):
    """
    MIMO class for the MRT scheme.

    The number of streams for the MRT scheme is always equal to one, but it
    still employs multiple transmit antennas.

    If `channel` is not provided you need to call the `set_channel_matrix`
    method before calling the other methods.

    Parameters
    ----------
    channel : np.ndarray
        MISO channel vector. It must be a 1D numpy array, where the
        number of receive antennas is assumed to be equal to 1.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        MisoBase.__init__(self, channel)

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:
        """
        Calculate the linear precoder for the MRT scheme.

        The MRT scheme corresponds to multiplying the symbol from each
        transmit antenna with a complex number corresponding to the inverse
        of the phase of the channel so as to ensure that the signals add
        constructively at the receiver. This also means that the MRT scheme
        only be applied to scenarios with a single receive antenna.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix with dimension (1, Nt).

        Returns
        -------
        W : np.ndarray
            The precoder that can be applied to the input data.
        """
        Nt = channel.shape[1]
        W = np.exp(-1j * np.angle(channel)).T / math.sqrt(Nt)
        return W

    @staticmethod
    def _calc_receive_filter(channel: np.ndarray,
                             noise_var: Optional[float] = None) -> np.ndarray:
        """
        Calculate the receive filter for the MRT scheme.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            The noise variance.

        Returns
        -------
        G_H : np.ndarray
            The receive_filter that can be applied to the input data.
        """
        Nt = channel.shape[1]
        G_H = math.sqrt(Nt) / np.sum(np.abs(channel))
        return G_H

    def encode(self, transmit_data: np.ndarray) -> np.ndarray:
        """
        Encode the transmit data array to be transmitted using the MRT
        scheme.

        The MRT scheme corresponds to multiplying the symbol from each
        transmit antenna with a complex number corresponding to the
        inverse of the phase of the channel so as to ensure that the
        signals add constructively at the receiver. This also means that
        the MRT scheme only be applied to scenarios with a single receive
        antenna.

        Parameters
        ----------
        transmit_data : np.ndarray
            A numpy array with the data to be transmitted.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.
        """
        # Add an extra first dimension so that broadcast does the right
        # thing later
        x = transmit_data[np.newaxis, :]
        W = self._calc_precoder(self._channel)

        # Element-wise multiplication (use broadcast)
        encoded_data = (W * x)
        return encoded_data

    def decode(self, received_data: np.ndarray) -> np.ndarray:
        """
        Decode the received data array.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the MRT scheme and
            corrupted by the channel `channel`.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        G_H = self._calc_receive_filter(self._channel)
        decoded_data = G_H * received_data
        decoded_data.shape = decoded_data.size

        return decoded_data


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx MRC Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class MRC(Blast):
    """
    MIMO class for the MRC scheme.

    The receive filter used will depend on the noise variance (see the
    :meth:`.set_noise_var` method). If the noise variance is positive the
    MMSE filter will be used, otherwise noise variance will be ignored and
    the Zero-Forcing filter will be used.

    The receive filter in the `Blast` class already does the maximum ratio
    combining. Therefore, this MRC class simply inherits from the Blast
    class and only exists for completion.

    Parameters
    ----------
    channel : np.ndarray
        MIMO channel matrix.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        Blast.__init__(self, channel)

    def set_channel_matrix(self, channel: np.ndarray) -> None:
        """
        Set the channel matrix.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix. The MRC MIMO scheme is defined for the
            scenario with multiple receive antennas and a single receive
            antenna. If channel is 1D assume that the number of transmit
            antennas is equal to 1.
        """
        # We will store the channel as a 2D numpy to be consistent with the
        # other MIMO classes
        if len(channel.shape) == 1:
            super(MRC, self).set_channel_matrix(channel[:, np.newaxis])
        else:
            super(MRC, self).set_channel_matrix(channel)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx SVD MIMO xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class SVDMimo(Blast):
    """
    MIMO class for the SVD MIMO scheme.

    Parameters
    ----------
    channel : np.ndarray
        MIMO channel matrix.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        Blast.__init__(self, channel)

    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:
        """
        Calculate the linear precoder for the SVD MIMO scheme.

        The SVD MIMO scheme employs as precoder the right singular matrix
        from SVD (Singular Value Decomposition) of the channel.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix with dimension (1, Nt).

        Returns
        -------
        W : np.ndarray
            The precoder that can be applied to the input data.
        """
        Nt = channel.shape[1]
        _, _, V_H = np.linalg.svd(channel)
        # The precoder is the 'V' matrix from the SVD decomposition of the
        # channel. Notice that we also need to divide by sqrt(Nt) to make
        # sure 'W' has a unitary norm.
        W = V_H.conj().T / math.sqrt(Nt)
        return W

    @staticmethod
    def _calc_receive_filter(channel: np.ndarray,
                             noise_var: Optional[float] = None) -> np.ndarray:
        """
        Calculate the receive filter for the SVD MIMO scheme.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            The noise variance.

        Returns
        -------
        G_H : np.ndarray
            The receive_filter that can be applied to the input data.
        """
        Nt = channel.shape[1]
        U, S, _ = np.linalg.svd(channel)
        G_H = np.diag(1. / S).dot(U.conj().T) * math.sqrt(Nt)
        return G_H

    def encode(self, transmit_data: np.ndarray) -> np.ndarray:
        """
        Encode the transmit data array to be transmitted using the SVD MIMO
        scheme.

        The SVD MIMO scheme corresponds to using the 'U' and 'V' matrices
        from the SVD decomposition of the channel as the precoder and
        receive filter.

        Parameters
        ----------
        transmit_data : np.ndarray
            A numpy array with the data to be transmitted.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.
        """
        num_elements = transmit_data.size
        if num_elements % self.Nt != 0:
            msg = ("Input array number of elements must be a multiple of the"
                   " number of transmit antennas")
            raise ValueError(msg)

        X = transmit_data.reshape(self.Nt, -1)

        W = self._calc_precoder(self._channel)
        encoded_data = W.dot(X)

        return encoded_data

    def decode(self, received_data: np.ndarray) -> np.ndarray:
        """
        Perform the decoding of the received_data for the SVD MIMO
        scheme with the channel `channel`.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the Alamouti scheme and
            corrupted by the channel `channel`.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        G_H = self._calc_receive_filter(self._channel)
        decoded_data = G_H.dot(received_data)

        # Return the decoded data as a 1D numpy array
        return decoded_data.reshape(decoded_data.size, )


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx GMD MIMO xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class GMDMimo(Blast):
    """
    MIMO class for the GMD based MIMO scheme.

    Parameters
    ----------
    channel : np.ndarray
        MIMO channel matrix.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        Blast.__init__(self, channel)

    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:
        """
        Calculate the linear precoder for the GMD scheme.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix with dimension (1, Nt).

        Returns
        -------
        W : np.ndarray
            The precoder that can be applied to the input data.
        """
        Nt = channel.shape[1]
        # The encode method will precode the transmit_data using the
        # matrix 'P' obtained from the gmd.
        U, S, V_H = np.linalg.svd(channel)
        _, _, P = gmd(U, S, V_H)
        W = P / math.sqrt(Nt)
        return W

    @staticmethod
    def _calc_receive_filter(channel: np.ndarray,
                             noise_var: Optional[float] = None) -> np.ndarray:
        """
        Calculate the receive filter for the MRT scheme.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.
        noise_var : float
            The noise variance.

        Returns
        -------
        G_H : np.ndarray
            The receive_filter that can be applied to the input data.
        """
        U, S, V_H = np.linalg.svd(channel)
        Q, R, _ = gmd(U, S, V_H)
        channel_eq = Q.dot(R)

        # Use the _calc_receive_filter method from the base class (Blast)
        G_H = Blast._calc_receive_filter(channel_eq, noise_var)
        return G_H

    def encode(self, transmit_data: np.ndarray) -> np.ndarray:
        """
        Encode the transmit data array to be transmitted using the GMD MIMO
        scheme.

        The GMD MIMO scheme is based on the Geometric Mean Decomposition
        (GMD) of the channel. The channel is decomposed into `H = Q R P^H`,
        where `R` is an upper triangular matrix with all diagonal elements
        being equal to the geometric mean of the singular values of the
        channel matrix `H`.

        corresponds to using the 'U' and 'V' matrices
        from the SVD decomposition of the channel as the precoder and
        receive filter.

        Parameters
        ----------
        transmit_data : np.ndarray
            A numpy array with the data to be transmitted.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.
        """
        num_elements = transmit_data.size
        if num_elements % self.Nt != 0:
            msg = ("Input array number of elements must be a multiple of the"
                   " number of transmit antennas")
            raise ValueError(msg)

        W = self._calc_precoder(self._channel)
        X = transmit_data.reshape(self.Nt, -1)
        encoded_data = W.dot(X)

        return encoded_data

    def decode(self, received_data: np.ndarray) -> np.ndarray:
        """
        Perform the decoding of the received_data for the GMD MIMO.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the Alamouti scheme and
            corrupted by the channel `channel`.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        G_H = self._calc_receive_filter(self._channel, self._noise_var)
        decoded_data = G_H.dot(received_data).reshape(-1)
        return decoded_data


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Alamouti Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class Alamouti(MimoBase):
    """
    MIMO class for the Alamouti scheme.

    Parameters
    ----------
    channel : np.ndarray
        MIMO channel matrix.
    """
    def __init__(self, channel: Optional[np.ndarray] = None):
        MimoBase.__init__(self, channel)

    @staticmethod
    def _calc_precoder(channel: np.ndarray) -> np.ndarray:  # pragma: nocover
        """
        Not defined.

        There is no linear precoder for the Almost scheme and thus an
        exception is called if this method is ever called.
        """
        raise RuntimeError("Alamouti scheme has no linear precoder")

    @staticmethod
    def _calc_receive_filter(
        channel: np.ndarray,
        noise_var: Optional[float] = None
    ) -> np.ndarray:  # pragma: nocover
        """
        Not defined.

        There is no linear receive filter for the Alamouti scheme that can
        be directly applied to the received data. Thus, an exception is
        called if this method is ever called.
        """
        raise RuntimeError("Alamouti scheme has no linear receive filter that"
                           " can be directly applied to the received data")

    def set_channel_matrix(self, channel: np.ndarray) -> None:
        """
        Set the channel matrix.

        Parameters
        ----------
        channel : np.ndarray
            MIMO channel matrix.

        Returns
        -------
        None
        """
        if len(channel.shape) == 1:
            super(Alamouti, self).set_channel_matrix(channel[np.newaxis, :])
        else:
            _, Nt = channel.shape
            if Nt != 2:
                msg = ("The number of transmit antennas must be equal to "
                       "2 for the {0} scheme").format(self.__class__.__name__)
                raise ValueError(msg)
            super(Alamouti, self).set_channel_matrix(channel)

    def getNumberOfLayers(self) -> int:
        """
        Get the number of layers of the Alamouti scheme.

        The number of layers in the Alamouti scheme is always equal to
        one.

        Returns
        -------
        Nl : int
            Number of layers of the Alamouti scheme, which is always one.
        """
        return 1

    def calc_linear_SINRs(self, noise_var: float) -> np.ndarray:
        """
        Calculate the SINRs (in linear scale) of the multiple streams.

        Parameters
        ----------
        noise_var : float
            The noise variance.

        Returns
        -------
        sinrs : np.ndarray
            The sinrs (in linear scale) of the multiple streams.
        """
        # The linear post-processing SINR for the Alamouti scheme is
        # given by
        # \[\frac{\Vert \mtH \Vert_F^2}{2 \sigma_N} \]
        sinr = np.linalg.norm(self._channel, 'fro')**2 / noise_var
        return sinr

    @staticmethod
    def _encode(transmit_data: np.ndarray) -> np.ndarray:
        """
        Perform the Alamouti encoding, but without dividing the power among
        the transmit antennas.

        The idea is that the encode method will call _encode and perform
        the power division. This separation allows better code reuse.

        Parameters
        ----------
        transmit_data : np.ndarray
            Data to be encoded by the Alamouti scheme.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data` (without dividing the power among
            transmit antennas).

        See also
        --------
        encode
        """
        Ns = transmit_data.size
        encoded_data = np.empty((2, Ns), dtype=complex)
        for n in range(0, Ns, 2):
            encoded_data[0, n] = transmit_data[n]
            encoded_data[0, n + 1] = -(transmit_data[n + 1]).conjugate()
            encoded_data[1, n] = transmit_data[n + 1]
            encoded_data[1, n + 1] = (transmit_data[n]).conjugate()
        return encoded_data

    def encode(self, transmit_data: np.ndarray) -> np.ndarray:
        """
        Perform the Alamouti encoding.

        Parameters
        ----------
        transmit_data : np.ndarray
            Data to be encoded by the Alamouti scheme.

        Returns
        -------
        encoded_data : np.ndarray
            The encoded `transmit_data`.
        """
        return self._encode(transmit_data) / math.sqrt(2)

    @staticmethod
    def _decode(received_data: np.ndarray, channel: np.ndarray) -> np.ndarray:
        """
        Perform the decoding of the received_data for the Alamouti
        scheme with the channel `channel`, but does not compensate for
        the power division among transmit antennas.

        The idea is that the decode method will call _decode and perform
        the power compensation. This separation allows better code reuse.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the Alamouti scheme and
            corrupted by the channel `channel`.
        channel : np.ndarray
            MIMO channel matrix.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data (without power compensating the power division
            performed during transmission).

        See also
        --------
        decode
        """
        Ns = received_data.shape[1]
        # Number of Alamouti codewords
        number_of_blocks = Ns // 2
        decoded_data = np.empty(Ns, dtype=complex)

        # Conjugate of the first column of the channel (first transmit
        # antenna to all receive antennas)
        h0_conj = channel[:, 0].conjugate()
        minus_h0 = -channel[:, 0]
        # Second column of the channel (second transmit antenna to all
        # receive antennas)
        h1 = channel[:, 1]
        h1_conj = channel[:, 1].conjugate()

        for i in range(0, number_of_blocks):
            decoded_data[2 * i] = (
                np.dot(h0_conj, received_data[:, 2 * i]) +
                np.dot(h1, received_data[:, 2 * i + 1].conjugate()))
            decoded_data[2 * i + 1] = (
                np.dot(h1_conj, received_data[:, 2 * i]) +
                np.dot(minus_h0, received_data[:, 2 * i + 1].conjugate()))

        # The Alamouti code gives a gain of the square of the Frobenius
        # norm of the channel. We need to compensate that gain.
        decoded_data /= np.linalg.norm(channel, 'fro')**2
        return decoded_data

    def decode(self, received_data: np.ndarray) -> np.ndarray:
        """
        Perform the decoding of the received_data for the Alamouti
        scheme with the channel `channel`.

        Parameters
        ----------
        received_data : np.ndarray
            Received data, which was encoded with the Alamouti scheme and
            corrupted by the channel `channel`.

        Returns
        -------
        decoded_data : np.ndarray
            The decoded data.
        """
        return self._decode(received_data, self._channel) * math.sqrt(2)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx



#!/usr/bin/env python

# pylint: disable=E1103
"""Module with implementation of channel related classes.
"""

__all__ = [
    'antennagain', 'fading_generators', 'fading', 'multiuser', 'noise',
    'pathloss', 'singleuser'
]



#!/usr/bin/env python
"""Module containing the simulation runner."""

import itertools
import os
import sys
from argparse import ArgumentParser
from time import time
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from pyphysim.simulations.progressbar import (ProgressBarBase,
                                              ProgressbarZMQClient)

from ..util.misc import pretty_time
from .parameters import SimulationParameters
from .progressbar import (ProgressBarIPython, ProgressbarText,
                          ProgressbarText2, ProgressbarText3,
                          ProgressbarZMQServer)
from .results import Result, SimulationResults

try:
    # noinspection PyUnresolvedReferences
    from ipyparallel import LoadBalancedView, DirectView
except ImportError:  # pragma: no cover
    LoadBalancedView = Any
    DirectView = Any

__all__ = ["get_partial_results_filename", "SimulationRunner", "SkipThisOne"]

UpdateFunction = Callable[[int], None]

ProxybarData = Tuple[int, str, int]

# A view form ipyparallel
ParallelView = Union[LoadBalancedView, DirectView]


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Command Line Argument Parser xxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def get_common_parser() -> ArgumentParser:
    """
    Get the command line parser that can be used.

    The global parser is a singleton object of the ArgumentParser class which is
    used in the `simulate_do_what_i_mean` function to parse command line
    arguments.

    It already has two arguments, "index" and "config" configured. If you need
    more then that, you can get the global parser object by calling this
    function and then calling the `add_argument` method of the returned object.
    See the documentation of ArgumentParser for more.

    Returns
    -------
    ArgumentParser
        The command line parser.
    """
    if get_common_parser.parser is None:  # type: ignore
        parser = ArgumentParser()
        group = parser.add_argument_group('General')

        help_msg = ('An index for the parameters variations. If provided, '
                    'only that variation will be simulated.')
        group.add_argument(
            '-i',  # short version to specify the opt
            '--index',  # Long version to specify the option
            help=help_msg,
            metavar='VARIATION INDEX',
            type=int,
            nargs='?')

        help_msg = 'Name of the file with the simulation parameters'
        group.add_argument(
            '-c',  # short version to specify the opt
            '--config',  # Long version to specify the opt
            help=help_msg,
            metavar='CONFIG FILENAME',
            # default=default_config_file,
            type=str,
            nargs='?')

        help_msg = ('Instead of running the simulation, return the '
                    'number of variations.')
        group.add_argument(
            # short version to specify the option
            '-n',
            # Long version to specify the option
            '--number_variations',
            help=help_msg,
            action='store_true')

        get_common_parser.parser = parser  # type: ignore

    return get_common_parser.parser  # type: ignore


get_common_parser.parser = None  # type: ignore

# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Module Functions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def get_partial_results_filename(
        results_base_filename: str,
        current_params: SimulationParameters,
        partial_results_folder: Optional[str] = None) -> str:
    """
    Get the name of the file where the partial result will be saved for
    the unpacked result with index `unpack_index`.

    Parameters
    ----------
    results_base_filename : str
        Base name for partial result file.
    current_params : SimulationParameters
        The current parameters, which must be a "unpacked variation" of
        another SimulationParameters object.
    partial_results_folder : str, optional
        The folder where the partial results will be stored.

    Returns
    -------
    partial_results_filename : str
        The name of the partial results file.
    """
    # This will get the number of unpacked variations of the parent
    # SimulationParameters object.
    total_unpacks = current_params.get_num_unpacked_variations()
    num_digits = len(str(total_unpacks))
    unpack_index_str = str(current_params.unpack_index).zfill(num_digits)

    partial_results_filename = '{0}_unpack_{1}.pickle'.format(
        results_base_filename, unpack_index_str)

    if partial_results_folder is not None:
        partial_results_filename = os.path.join(partial_results_folder,
                                                partial_results_filename)

    return partial_results_filename


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Exception xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class SkipThisOne(Exception):
    """
    Exception that can be raised in the `_run_simulation` method to skip
    the current repetition.

    The `simulate` method will not count a run of `_run_simulation` if it
    throws a `SkipThisOne` exception.

    Parameters
    ----------
    msg : str
        The message with more information on why the exception was
        raised.
    """
    def __init__(self, msg: str) -> None:
        """
        Parameters
        ----------
        msg : str
            The message with more information on why the exception was
            raised.
        """
        super().__init__()
        self.msg = msg

    def __str__(self) -> str:  # pragma: nocover
        """
        Convert the exception object to a suitable string representation.

        Returns
        -------
        str
            String representation of the object.
        """
        return "SkipThisOne: {0}".format(self.msg)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx SimulationRunner - START xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# pylint: disable=R0902
class SimulationRunner:
    """
    Base class to run Monte Carlo simulations.

    The main idea of the :class:`SimulationRunner` class is that in order
    to implement a Monte Carlo simulation one would subclass
    :class:`SimulationRunner` and implement the :meth:`_run_simulation`
    method (as well as any of the optional methods). The complete procedure
    is described in the documentation of the :mod:`simulations` module.

    The code below illustrates the minimum pseudo code to implement a
    subclass of :class:`SimulationRunner`.

    .. code-block:: python

       class SomeSimulator(SimulationRunner):
       def __init__(self):
           SimulationRunner.__init__(self)
           # Do whatever you need/want

           # Add the simulation parameters to the `params` attribute.
           self.params.add('par1', par1value)
           ...
           self.params.add('parN', parNvalue)
           # Optionally set some parameter(s) to be unpacked
           self.params.set_unpack_parameter('name_of_a_parameter')

       def _run_simulation(self, current_parameters):
           # Get the simulation parameters from the current_parameters
           # object. If no parameter was marked to be unpacked, then
           # current_parameters will be equivalent to self.params.
           par1 = current_parameters['par1']
           ...
           parN = current_parameters['parN']

           # Do the simulation of one iteration using the parameters
           # par1,...parN from the current_parameters object.
           ...

           # Save the results of this iteration to a SimulationResults
           # object and return it
           simResults = SimulationResults()
           simResults.add_new_result(...)  # Add one each result you want
           simResults.add_new_result(...)
           return simResults

    With that, all there is left to run the simulation is to create a
    SomeSimulator object and call its :meth:`simulate` method.

    Parameters
    ----------
    default_config_file : str
        Name of the config file. This will be parsed with configobj.
    config_spec : list[str]
        Configuration specification to validate the config file.
    read_command_line_args : bool
        If True (default), read and parse command line arguments.
    save_parsed_file : bool
        If True, the config file will be saved after it is loaded. This is
        useful to add the default parameters to the config file so that
        they can be easily changed later. Note that if the config file does
        not exist at all, then it will be saved regardless of the value of
        `save_parsed_file`.

    See Also
    --------
    .SimulationResults : Class to store simulation results.
    .SimulationParameters : Class to store the simulation parameters.
    .Result : Class to store a single simulation result.
    """
    def __init__(self,
                 default_config_file: Optional[str] = None,
                 config_spec: Optional[str] = None,
                 read_command_line_args: bool = True,
                 save_parsed_file: bool = False) -> None:
        self.rep_max = 1
        # Number of iterations performed by simulation when it finished
        self._runned_reps: List[int] = []

        self._config_filename = None
        # Configobj specification (to validate parameters read from the
        # config file)
        self._configobj_spec = config_spec

        # xxxxx Parse command line arguments (get config filename) xxxxxxxx
        if read_command_line_args is True:  # pragma: no cover
            # Note that the get_common_parser always return the same object
            parser = get_common_parser()

            # This member variable will store all the parsed command line
            # arguments
            self.command_line_args = parser.parse_args()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            # Get the config filename if it was provided in the command
            # line. If not, use the default value.
            if self.command_line_args.config is None:
                self._config_filename = default_config_file
            else:  # pragma: no cover
                self._config_filename = self.command_line_args.config
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        else:  # pragma: no cover
            # Since we are not parsing command line arguments, the config
            # file will be the provided default_config_file
            self._config_filename = default_config_file
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxx Read the parameters from the config file xxxxxxxxxxxxx
        if self._config_filename is None:
            self.params = SimulationParameters()
        else:  # pragma: no cover
            if not os.path.isfile(self._config_filename):
                # If the config file does not exist, we will save the file
                # no matter the value of save_parsed_file
                save_parsed_file = True

            self.params = SimulationParameters.load_from_config_file(
                self._config_filename,
                self._configobj_spec,
                save_parsed_file=save_parsed_file)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        self.results = SimulationResults()

        # This variable will be used later to store the progressbar object
        # when it is created in the _get_update_progress_function method
        self._pbar: Optional[ProgressBarBase] = None

        # xxxxxxxxxx update_progress_function_style xxxxxxxxxxxxxxxxxxxxxxx
        # --- When the simulation is performed Serially -------------------
        # Sets the style of the used progressbar. The allowed values are
        # 'text1', 'text2', None, or a callable object.
        # - If it is 'text1' then the ProgressbarText class will be used.
        # - If it is 'text2' then the ProgressbarText2 class will be used.
        # - If it is None, then no progressbar will be used.
        # - If it is a callable, then that callable object must receive two
        #   arguments, the rep_max and the message values, and return a
        #   function that receives a single argument (the custom
        #   parameters).
        # --- When the simulation is performed in parallel ----------------
        # - If it is None then no progressbar will be used
        # - If it is not None then a socket progressbar will be used, which
        #   employs the same style as 'text2'
        self.update_progress_function_style: Optional[str] = 'text2'

        # This can be either 'screen' or 'file'. If it is 'file' then the
        # progressbar will write the progress to a file with appropriated
        # filename
        self.progress_output_type = 'screen'

        # Dictionary with extra arguments that will be passed to the
        # __init__ method of the progressbar class. For instance, when
        # simulating in parallel and update_progress_function_style is not
        # None a progressbar based on ZMQ sockets will be used. Set
        # progressbar_extra_args to "{'port':3456}" in order for the
        # progressbar to use port 3456.
        self.progressbar_extra_args: Dict[str, Any] = {}

        # Additional message printed in the progressbar. The message can
        # contain "{SomeParameterName}" which will be replaced with the
        # parameter value.
        #
        # Note that if the update_progress_function_style is None, then no
        # message will be printed either.
        self.progressbar_message = 'Progress'

        # This variable will be used to store the AsyncMapResult object
        # that will be created in the simulate_in_parallel method. This
        # object is part of ipyparallel framework and is used to get
        # the actual results of performing an asynchronous task in IPython.
        self._async_results: Optional[Any] = None

        # xxxxxxxxxx Configure saving of simulation results xxxxxxxxxxxxxxx
        # If this variable is set to True the saved partial results will be
        # deleted after the simulation is finished.
        self.delete_partial_results_bool = False

        # Folder where the partial results will be saved. Set this to None
        # to save the partial results in the same folder of the final
        # results.
        self.partial_results_folder = 'partial_results'
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxx Internal variables you should not modify xxxxxxxxxxxxxxxxxx
        # Variable to store the name of the file where the simulation
        # results will be stored.
        self._results_base_filename: Optional[str] = None
        # Variable to store all the names for the partial results. Each
        # name in it will be equivalent to the value of
        # _results_base_filename appended with unpack_i where i will be an
        # integer. These names will be used after the simulation has
        # finished and full results were saved to delete the files with the
        # partial results.
        self._results_base_filename_unpack_list: List[str] = []
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxx Interval variables for tracking simulation time xxxxxxxxxxx
        # Note that self._elapsed_time is different from the 'elapsed_time'
        # result returned after the simulation has finished. This variable
        # only tracks the CURRENT SIMULATION and does not account the time
        # any loaded partial results required to be simulated.
        self._elapsed_time = 0.0
        self.__tic = 0.0
        self.__toc = 0.0
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def set_results_filename(self, filename: Optional[str] = None) -> None:
        """
        Set the name of the file where the simulation results will be
        saved.

        This must be done before calling the `simulate` of the
        `simulate_in_parallel` methods.

        The `filename` argument is formatted with the simulation
        parameters. That is, suppose there are two parameters Nr=2 and
        Nt=1, then if `filename` is equal to "results_for_{Nr}x{Nt}" the
        actual name of the file used to store the simulation results will
        be "results_for_2x1.pickle".

        The advantage of setting the name of the results file with
        `set_results_filename` instead of manually saving the results after
        the simulation is finished is that partial results will also be
        saved. Therefore the simulation can be stopped and continued later
        from these partial results.

        Parameters
        ----------
        filename : str
            The name of the file (without extension) where the simulation
            results will be stored. If not provided the results will not be
            automatically stored.
        """
        self._results_base_filename = filename

    @property
    def results_base_filename(self) -> Optional[str]:
        """
        Get name of the file where the last simulation results were stored.

        Returns
        -------
        str
            The result base filename.
        """
        if self._results_base_filename is None:
            filename = None
        else:
            # noinspection PyTypeChecker
            filename = self.results.get_filename_with_replaced_params(
                self._results_base_filename)
        return filename

    @property
    def results_filename(self) -> Optional[str]:
        """
        Get name of the file where the last simulation results were stored.

        Returns
        -------
        str
            The result base filename.
        """
        results_base_filename = self.results_base_filename
        if results_base_filename is None:
            results_filename = None
        else:
            ext = os.path.splitext(results_base_filename)[-1]
            if ext == '.pickle':
                results_filename = results_base_filename
            else:
                results_filename = '{0}.pickle'.format(results_base_filename)

        return results_filename

    def _get_progress_output_sink(
            self,
            param_variation_index: Union[int, str]) -> Any:  # pragma: no cover
        """
        Get the output sink for the progressbars.

        If 'self.progress_output_type' is equal to 'screen' this method
        will simple return sys.stdout.

        Parameters
        ----------
        param_variation_index : int | str
            Int or a string that can be converted to int. If this is
            provided and 'self.progress_output_type' is 'file',
            this will be used in the filename of the progress output file.

        Returns
        -------
        out : sys.stdout or a file object
        """
        out = sys.stdout
        if self.progress_output_type == 'screen':
            pass
        else:
            total_unpacks = self.params.get_num_unpacked_variations()
            num_digits = len(str(total_unpacks))
            unpack_index_str = str(param_variation_index).zfill(num_digits)
            filename = '{0}_progress_{1}_of_{2}.txt'.format(
                self.results_base_filename, unpack_index_str, total_unpacks)
            out = open(filename, 'w')

        return out

    def __delete_partial_results_maybe(self) -> None:
        """
        (maybe) Delete the files containing partial results.

        This method is called inside the simulate method after the full
        results were saved. It will only have an effect if the
        "delete_partial_results_bool" variable is True (thus the 'maybe' in
        the method's name).

        Notes
        -----
        This method will do nothing if self.delete_partial_results_bool is
        not True.
        """
        if self.delete_partial_results_bool is True:
            for name in self._results_base_filename_unpack_list:
                try:
                    os.remove(name)
                except OSError:  # pragma: no cover
                    pass
            self._results_base_filename_unpack_list = []
        else:
            # Do nothing if self.delete_partial_results_bool is not True
            pass

    def __save_partial_results(self, current_rep: int,
                               current_params: SimulationParameters,
                               current_sim_results: SimulationResults,
                               partial_results_filename: str) -> None:
        """
        Save the partial simulation results to a file.

        Parameters
        ----------
        current_rep : int
            Current repetition.
        current_params : SimulationParameters
            The current parameters.
        current_sim_results : SimulationResults
            The partial simulations results object to be saved.
        partial_results_filename : str
            The name of the file to save the partial simulation results.

        Notes
        -----
        The name of the results file where the partial results were saved,
        that is, the value of the `partial_results_filename` argument will
        be automatically added to the list of files to be deleted after the
        simulation is finished if delete_partial_results_bool is
        True. However, remember that in the simulate_in_parallel method
        this method will be run in a different object in an IPython
        engine. Therefore, you will need to manually add, the value of the
        partial_results_filename variable to the list of files to be
        deleted (the __results_base_filename_unpack_list variable).
        """
        # xxxxxxxxxx Save partial results to file xxxxxxxxxxxxxxxxxxxxx
        # First we add the current parameters to the partial simulation
        # results object
        current_sim_results.set_parameters(current_params)
        # Now we can save the partial results to a file.
        current_sim_results.current_rep = current_rep

        # Try to save the partial results
        try:
            filename = current_sim_results.save_to_file(
                partial_results_filename)
        except IOError as e:
            if self.partial_results_folder is not None:

                os.mkdir(self.partial_results_folder)
                # This should not raise IOError again.
                filename = current_sim_results.save_to_file(
                    partial_results_filename)
            else:  # pragma: no cover
                raise e

        self._results_base_filename_unpack_list.append(filename)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def clear(self) -> None:  # pragma: no cover
        """
        Clear the SimulationRunner.

        This will erase any results from previous simulations as well as
        other internal variables. The SimulationRunner object will then be
        as if it was just created, except that the simulation parameters
        will be kept.

        """
        self._elapsed_time = 0.0
        # Number of iterations performed by simulation when it finished
        self._runned_reps = []
        self.results = SimulationResults()

    def __run_simulation_and_track_elapsed_time(
            self,
            current_parameters: SimulationParameters) -> SimulationResults:
        """
        Perform the _run_simulation method and track its execution time.
        This time will be added as a Result to the returned
        :class:`.SimulationResults` object from _run_simulation.

        Parameters
        ----------
        current_parameters : SimulationParameters
            SimulationParameters object with the parameters for the
            simulation. The self.params variable is not used directly. It
            is first unpacked in the simulate function which then calls
            _run_simulation for each combination of unpacked parameters.

        Returns
        -------
        SimulationResults
            The current simulation results.

        Notes
        -----
        This method is called in the `simulate` and `simulate_in_parallel`.
        """
        tic = time()
        current_sim_results = self._run_simulation(current_parameters)
        toc = time()
        elapsed_time_result = Result.create('elapsed_time', Result.SUMTYPE,
                                            toc - tic)
        current_sim_results.add_result(elapsed_time_result)

        return current_sim_results

    def _run_simulation(
            self,
            current_parameters: SimulationParameters) -> SimulationResults:
        """
        Performs one iteration of the simulation.

        This function must be implemented in a subclass. It should take the
        needed parameters from the params class attribute (which was filled
        in the constructor of the derived class) and return the results as
        a :class:`.SimulationResults` object.

        Note that _run_simulation will be called self.rep_max times (or
        less if an early stop criteria is reached, which requires
        reimplementing the _keep_going function in the derived class) and
        the results from multiple repetitions will be merged.

        Parameters
        ----------
        current_parameters : SimulationParameters
            SimulationParameters object with the parameters for the
            simulation. The self.params variable is not used directly. It
            is first unpacked in the simulate function which then calls
            _run_simulation for each combination of unpacked parameters.

        Returns
        -------
        simulation_results : SimulationResults
            A SimulationResults object containing the simulation results of
            the run iteration.

        """
        raise NotImplementedError("'_run_simulation' must be implemented "
                                  "in a subclass of SimulationRunner")

    # pylint: disable=W0613,R0201
    def _keep_going(self, current_params: SimulationParameters,
                    current_sim_results: SimulationResults,
                    current_rep: int) -> bool:
        """
        Check if the simulation should continue or stop.

        This function may be reimplemented in the derived class if a stop
        condition besides the number of iterations is desired. The idea is
        that _run_simulation returns a :class:`.SimulationResults` object, which is
        then passed to _keep_going, which is then in charge of deciding if
        the simulation should stop or not.

        Parameters
        ----------
        current_params : SimulationParameters
            SimulationParameters object with the parameters of the
            simulation.
        current_sim_results : SimulationResults
            SimulationResults object from the last iteration (merged with
            all the previous results)
        current_rep : int
            Number of iterations already run.

        Returns
        -------
        result : bool
            True if the simulation should continue or False otherwise.

        Notes
        -----
        This method should be simple (it will be run many times) and SHOULD
        NOT modify the object.
        """
        # If this function is not reimplemented in a subclass it always
        # returns True. Therefore, the simulation will only stop when the
        # maximum number of allowed iterations is reached.
        return True

    def _get_serial_update_progress_function(
        self, current_params: SimulationParameters
    ) -> UpdateFunction:  # pragma: no cover
        """
        Return a function that should be called to update the
        progressbar for the simulation of the current parameters.

        This method is only called in the 'simulate' method.

        The returned function accepts a single argument, corresponding to
        the number of iterations executed so far.

        The progressbar used to get the returned function depend on the
        value of the self.update_progress_function_style attribute.

        Parameters
        ----------
        current_params : SimulationParameters
            The current combination of simulation parameters. This should
            be used to perform any replacement in the
            self.progressbar_message string that will be written in the
            progressbar. Note that string replacement will also work in
            self.progressbar_message for a 'rep_max' field.

        Returns
        -------
        func : (int) -> []
            A function that accepts a single integer argument and can be
            called to update the progressbar.

        Notes
        -----
        The equivalent of this method which is used in the
        simulate_in_parallel method if the
        _get_parallel_update_progress_function method.
        """
        if 'rep_max' not in current_params.parameters:
            current_params.parameters['rep_max'] = self.rep_max

        # If the progressbar_message has any string replacements in the
        # form {some_param} where 'some_param' is a parameter in
        # current_params then it will be replaced by the current value of
        # 'some_param'.
        message = self.progressbar_message.format(**current_params.parameters)

        # By default, the returned function is a dummy function that does
        # nothing
        def update_progress_func(_: int) -> None:
            pass

        if self.update_progress_function_style is None:
            return update_progress_func

        # If self.progress_output_type is equal to 'screen', this will be
        # sys.stdout, otherwise it will be a file object. Note that
        # self.progress_output_type is used when
        # self.update_progress_function_style is either 'text1' or 'text2'.
        output_progress_sink = self._get_progress_output_sink(
            current_params.unpack_index)

        # If the self.update_progress_function_style attribute matches one
        # of the available styles, then update_progress_func will be
        # appropriately set.
        if self.update_progress_function_style == 'text1':  # pragma: no cover
            # We will use the ProgressbarText class
            # noinspection PyArgumentList
            self._pbar = ProgressbarText(self.rep_max,
                                         '*',
                                         message,
                                         output=output_progress_sink,
                                         **self.progressbar_extra_args)
            update_progress_func = self._pbar.progress
            self._pbar.delete_progress_file_after_completion = True
        elif self.update_progress_function_style == 'text2':
            # We will use the ProgressbarText2 class
            # noinspection PyArgumentList
            self._pbar = ProgressbarText2(self.rep_max,
                                          '*',
                                          message,
                                          output=output_progress_sink,
                                          **self.progressbar_extra_args)
            update_progress_func = self._pbar.progress
            self._pbar.delete_progress_file_after_completion = True
        elif self.update_progress_function_style == 'ipython':
            self._pbar = ProgressBarIPython(self.rep_max, message)
            update_progress_func = self._pbar.progress

        elif callable(self.update_progress_function_style) is True:
            # We will use a custom function to update the progress. Note
            # that we call self.update_progress_function_style to return
            # the actual function that will be used to update the
            # progress. That is, the function stored in
            # self.update_progress_function_style should basically do what
            # _get_update_progress_function is supposed to do.

            # pylint: disable=E1102
            # noinspection PyCallingNonCallable
            update_progress_func = self.update_progress_function_style(
                self.rep_max, self.progressbar_message)  # pragma: no cover

        return update_progress_func

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def _get_parallel_update_progress_function(
            self) -> ProxybarData:  # pragma: no cover
        """
        Return a function that should be called to update the
        progressbar for the simulation of the current parameters.

        This method is only called in the 'simulate_in_parallel' method.

        The returned function accepts a single argument, corresponding to
        the number of iterations executed so far.

        # The progressbar used to get the returned function depend on the
        # value of the self.update_progress_function_style attribute.

        Returns
        -------
        list[int,int]
            List with the proxybar client_id, ip and port.

        Notes
        -----
        While the _get_serial_update_progress_function method receives the
        "current parameters" this _get_parallel_update_progress_function
        receives the full parameters. The reason for this is that when the
        simulation is performed in parallel multiple process (for different
        parameters) will update the same progressbar. Therefore, it does
        not make sense to perform replacements in the progressbar message
        based on current parameters.
        """
        if self.update_progress_function_style is not None:
            if self._pbar is None:
                parameters = self.params.parameters
                # If the progressbar_message has any string replacements
                #  in the form {some_param} where 'some_param' is a
                # parameter in 'full_params' then it will be replaced by
                #  the value of 'some_param'.
                message = self.progressbar_message.format(**parameters)

                sleep_time = 1.0
                if self.progress_output_type == 'screen':
                    filename = None
                else:
                    # Lets update the progress every 30 seconds
                    sleep_time = 30.0
                    filename = '{0}_progress.txt'.format(
                        self._results_base_filename)

                self._pbar = ProgressbarZMQServer(
                    message=message,
                    sleep_time=sleep_time,
                    filename=filename,
                    **self.progressbar_extra_args)

            # Note that this will be an object of the ProgressbarZMQClient
            # class, but it behaves like a function.
            proxybar = \
                self._pbar.register_client_and_get_proxy_progressbar(
                    self.rep_max)
            proxybar_data = [proxybar.client_id, proxybar.ip, proxybar.port]
        else:
            return None

        return proxybar_data

    @property
    def elapsed_time(self) -> float:
        """
        Get the simulation elapsed time. Do not set this value.

        Returns
        -------
        float
            The elapsed time.
        """
        return pretty_time(self._elapsed_time)

    @property
    def runned_reps(self) -> List[int]:
        """
        Get method for the runned_reps property.

        Returns
        -------
        list[int]
            The value of the runned_reps property, which stores the
            number of runned repetitions for each unpacked parameters
            combination.
        """
        return self._runned_reps

    # This method is called when the SimulationRunner class is pickled
    def __getstate__(self):  # pragma: no cover
        # We will pickle everything as default, except for the "_pbar"
        # member variable that will not be pickled. The reason is that
        # it may be a ProgressbarZMQServer object, which cannot be
        # pickled (uses ZMQ sockets).
        state = dict(self.__dict__)
        del state['_pbar']
        return state

    # def get_runned_reps_fix_params(
    #         self,
    #         fixed_params_dict=None):  # pragma: no cover
    #     """
    #     Get the number of runned repetitions for a given set of parameters.

    #     You can get a list of the number of repetitions combination of
    #     transmit parameters with the "runned_reps" property. However, if
    #     you have more then one transmit parameter set to be unpacked it
    #     might be difficult knowing which element in the list corresponds to
    #     the simulation for a given set of transmit parameters. By using the
    #     get_runned_reps_fix_params method you will be the number
    #     repetitions for the desired set of transmit parameters.

    #     Parameters
    #     ----------
    #     fixed_params_dict : dictionary
    #     """
    #     if fixed_params_dict is None:
    #         fixed_params_dict = dict()

    #     indexes = self.params.get_pack_indexes(fixed_params_dict)
    #     runned_reps_subset = np.array(self.runned_reps)[indexes]
    #     return runned_reps_subset

    # noinspection PyUnboundLocalVariable
    def _simulate_for_current_params_common(
        self,
        current_params: SimulationParameters,
        update_progress_func: UpdateFunction = lambda value: None
    ) -> Tuple[int, SimulationResults, str]:  # pragma: no cover
        """
        Parameters
        ----------
        current_params : SimulationParameters
            The current parameters
        update_progress_func : (int) -> []
            The function that can be called to update the current progress.

        Returns
        -------
        (int, SimulationResults, str)
            The value of `current_rep`, the current results as a
            SimulationResults object, and the name of the file storing
            partial results.
        """
        # Implement the _on_simulate_current_params_start method in a
        # subclass if you need to run code before the _run_simulation
        # iterations for each combination of simulation parameters.
        self._on_simulate_current_params_start(current_params)

        # First we try to Load the partial results for the current
        # parameters.
        try:
            # Name of the file where the partial results will be saved
            if self._results_base_filename is not None:
                partial_results_filename = get_partial_results_filename(
                    self.results_base_filename, current_params,
                    self.partial_results_folder)
            else:
                # If __results_base_filename is None there is also no
                # partial results to load. Therefore, lets raise an
                # IOError here to go to the except catching
                raise IOError()

            # If loading partial results succeeds, then we will have
            # partial results. If it fails because the file does not
            # exist, this will thrown a IOError exception and we will
            # execute the except block instead.
            current_sim_results = SimulationResults.load_from_file(
                partial_results_filename)
            num_skipped_reps_result = Result.create("num_skipped_reps",
                                                    Result.SUMTYPE, 0)
            current_sim_results.add_result(num_skipped_reps_result)

            # Note that at this point we have successfully loaded the
            # partial results from the file. However, we still need to
            # make sure it matches our current parameters. It it does
            # not match the user likely used a wrong name and we raise
            # an exception to stop the simulation.
            #
            # NOTE: If the type of this exception is changed in the
            # future make sure it is not caught in the except block.
            if not current_params == current_sim_results.params:
                err_msg = ("Partial results loaded from file does not match"
                           " current parameters. \nfile: {0}")
                raise ValueError(err_msg.format(partial_results_filename))

            # The current_rep will be set to the value or run
            # repetitions in the loaded partial results. This means
            # that the "while" statement after this try/except block
            # will have a head start and if current_rep is greater than
            # or equal to rep_max the while loop won't run at all.
            # noinspection PyUnresolvedReferences
            current_rep = current_sim_results.current_rep

        # If loading partial results failed then we will run the FIRST
        # repetition here and the "while" statement after this
        # try/except block will run as usual.
        except IOError:
            # Perform the first iteration of _run_simulation
            current_sim_results = \
                self.__run_simulation_and_track_elapsed_time(
                    current_params)
            # Add the extra 'num_skipped_reps' Result.
            current_sim_results.add_new_result('num_skipped_reps',
                                               Result.SUMTYPE, 0)
            current_rep = 1

        last_tic = time()
        # Run more iterations until one of the stop criteria is
        # reached. Note that if partial results were loaded successfully
        # from file and they already achieve the stop criteria then the
        # while loop below will not run.
        while (self._keep_going(current_params, current_sim_results,
                                current_rep) and current_rep < self.rep_max):
            # xxxxxxxxxx Run one repetition of the simulation xxxxxxxxxxxxx
            try:
                # Run one repetition of the `_run_simulation` and merge the
                # new results. If `_run_simulation` raises a SkipThisOne
                # exception, then we do not increase current_rep or the
                # current progress, since there is no new result to merge.
                current_sim_results.merge_all_results(
                    self.__run_simulation_and_track_elapsed_time(
                        current_params))

                current_rep += 1
                update_progress_func(current_rep)
            except SkipThisOne:
                # Each time a SkipThisOne exception is raised we increase
                # the num_skipped_reps_reps result to indicate that, but we
                # don't increase the current progress or the
                # current_rep. After that, the while loop will continue the
                # simulation.
                current_sim_results['num_skipped_reps'][-1].update(1)
                # TODO: Maybe log that one repetition was skipped
                # print("\nAlready skipped {0} repetitions").format(
                #     current_sim_results['num_skipped_reps'][-1].get_result())
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            toc = time()
            # Save partial results each 500 iterations as well as each 5
            # minutes
            if ((toc - last_tic > 300 or current_rep % 500 == 0)
                    and self._results_base_filename is not None):
                self.__save_partial_results(current_rep, current_params,
                                            current_sim_results,
                                            partial_results_filename)
                last_tic = time()

        # If the while loop ended before rep_max repetitions (because
        # _keep_going returned false) then set the progressbar to full.
        update_progress_func(self.rep_max)

        # Implement the _on_simulate_current_params_finish method in a
        # subclass if you need to run code after all _run_simulation
        # iterations for each combination of simulation parameters
        # finishes.
        self._on_simulate_current_params_finish(current_params,
                                                current_sim_results)

        # xxxxxxxxxx Save partial results to file xxxxxxxxxxxxxxxxxxxxx
        # Save partial results for current parameters after all repetitions
        if self._results_base_filename is not None:
            self.__save_partial_results(current_rep, current_params,
                                        current_sim_results,
                                        partial_results_filename)
        else:
            partial_results_filename = None
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # This function returns a tuple containing the number of
        # iterations run as well as the SimulationResults object.
        return current_rep, current_sim_results, partial_results_filename

    def _simulate_for_current_params_serial(
        self, current_params: SimulationParameters,
        var_print_iter: Iterator[int]
    ) -> Tuple[int, SimulationResults, str]:
        """
        Simulate (serial) for the current parameters.

        Parameters
        ----------
        current_params : SimulationParameters
            The current parameters
        var_print_iter : iterator
            An iterator obtained from calling the
            __get_print_variation_iterator method.

        Returns
        -------
        (int, SimulationResults, str)
            The value of `current_rep`, the current results as a
            SimulationResults object, and the name of the file storing
            partial results.
        """
        # (maybe) Print the current variation. This is something like
        # ------------- Current Variation: 4/84 ------------
        next(var_print_iter)

        update_progress_func = self._get_serial_update_progress_function(
            current_params)

        return self._simulate_for_current_params_common(
            current_params, update_progress_func)

    # This method is run in another process. Therefore, the python coverage
    # program cannot see that it is actually used (it is used when the
    # simulate_in_parallel method is tested). Because of that we add the
    # pragma line here.
    @staticmethod
    def _simulate_for_current_params_parallel(
        obj: "SimulationRunner",
        current_params: SimulationParameters,
        proxybar_data=None
    ) -> Tuple[int, SimulationResults, str]:  # pragma: no cover
        """
        Simulate (parallel) for the current parameters.

        Parameters
        ----------
        obj : SimulationRunner
            The same as the self parameter in regular methods. The reason
            that this method is set to static is to allow it to be pickled.
        current_params : SimulationParameters
            The current parameters
        proxybar_data : (int,str,int) | None
                The elements are the "client_id" (and int), the "ip" (a
                string with an IP address) and the "port". This data should
                be used to create a ProgressbarZMQClient object that can be
                used to update the progressbar (via a ZMQ socket)

        Returns
        -------
        (int, SimulationResults, str)
            The value of `current_rep`, the current results as a
            SimulationResults object, and the name of the file storing
            partial results.
        """
        # xxxxxxxxxx Function to update the progress xxxxxxxxxxxxxxxxxx
        if proxybar_data is None:

            def update_progress_func(_):
                pass
        else:
            client_id, ip, port = proxybar_data  # pylint: disable=W0633
            proxybar = ProgressbarZMQClient(client_id, ip, port)
            update_progress_func = proxybar.progress
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # pylint: disable= W0212
        # noinspection PyProtectedMember
        return obj._simulate_for_current_params_common(current_params,
                                                       update_progress_func)

    def __get_print_variation_iterator(self,
                                       num_variations: int,
                                       start: int = 0) -> Iterator[int]:
        """
        Returns an iterator that prints the current variation each time
        it's "next" method is called.

        Create the var_print_iter Iterator Each time the 'next' method of
        var_print_iter is called it will print something like

        ------------- Current Variation: 4/84 ------------

        which means the variation 4 of 84 variations.

        Parameters
        ----------
        num_variations : int
            The number of different variations.
        start : int (default is 1)
            The index of the first variation.

        Returns
        -------
        iterator
            An iterator that can be called to print the current variation.
        """
        if (self.update_progress_function_style is None
                or self.progress_output_type != 'screen'
                or self.update_progress_function_style == 'ipython'):
            for _ in itertools.repeat(''):
                yield 0
        else:  # pragma: no cover
            variation_pbar = ProgressbarText3(num_variations,
                                              progresschar='-',
                                              message="Current Variation:")

            for i in range(start, num_variations):
                variation_pbar.progress(i + 1)
                print()  # print a new line
                yield i

    def simulate(self, param_variation_index: Optional[int] = None) -> None:
        """
        Performs the full Monte Carlo simulation (serially).

        Implements the general code for every simulation. Any code
        specific to a single simulation iteration must be implemented in
        the _run_simulation method of a subclass of SimulationRunner.

        The main idea behind the SimulationRunner class is that the general
        code in every simulator is implemented in the SimulationRunner
        class, more specifically in the `simulate` method, while the
        specific code of a single iteration is implemented in the
        _run_simulation method in a subclass.

        Parameters
        ----------
        param_variation_index : int, optional
            If not provided, the full simulation will be run. If this is
            provided the simulation will be run only for the parameters
            variation with index given by `param_variation_index`. In that
            case, calling the set_results_filename method before the
            simulate method is required since only the partial results will
            be saved.

        See Also
        --------
        simulate_in_parallel
        """
        if param_variation_index is not None:  # pragma: no cover
            # Maybe even though param_variation_index is a valid integer it
            # was passed as a string. Let's try to convert whatever we have
            # to an integer.
            param_variation_index = int(param_variation_index)

        # xxxxxxxxxxxxxxx Some initialization xxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        self.__tic = time()
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxx Store rep_max in the results object xxxxxxxxxxxxxxxxxxxxxxx
        self.results.rep_max = self.rep_max
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Store the Simulation parameters in the SimulationResults object.
        # With this, the simulation parameters will be available for
        # someone that has the SimulationResults object (loaded from a
        # file, for instance).
        self.results.set_parameters(self.params)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Implement the _on_simulate_start method in a subclass if you need
        # to run code at the start of the simulate method.
        self._on_simulate_start()
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Get the number of variations of the transmit parameters
        num_variations = self.params.get_num_unpacked_variations()

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxx Start of the code unique to the serial version xxxxxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Create the var_print_iter Iterator
        # Each time the 'next' method of var_print_iter is called it will
        # print something like
        # ------------- Current Variation: 4/84 ------------
        # which means the variation 4 of 84 variations.

        # However, this will only be printed if
        # self.progress_output_type is equal to 'screen'
        if param_variation_index is None:
            var_print_iter = self.__get_print_variation_iterator(
                num_variations)
        else:
            var_print_iter = self.__get_print_variation_iterator(
                num_variations, start=param_variation_index)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Here we can either simulate for a single combination of
        # parameters, if param_variation_index was provided, or for all
        # combinations if it is not provided.
        if param_variation_index is not None:
            if self._results_base_filename is None:
                err_msg = ('The results filename must be set before'
                           ' calling the "simulate" method.')
                raise RuntimeError(err_msg)
            param_comb_list = self.params.get_unpacked_params_list()

            if 0 <= param_variation_index < len(param_comb_list):
                current_params = param_comb_list[param_variation_index]
                self._simulate_for_current_params_serial(
                    current_params, var_print_iter)

        # If param_variation_index is None we will run for all parameters
        # combinations
        else:
            # xxxxx FOR UNPACKED PARAMETERS xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            # Loop through all the parameters combinations
            for current_params in self.params.get_unpacked_params_list():
                (current_rep, current_sim_results, _) \
                    = self._simulate_for_current_params_serial(
                        current_params, var_print_iter)

                # Store the number of repetitions actually ran for the
                # current parameters combination
                self._runned_reps.append(current_rep)
                # Lets append the simulation results for the current
                # parameters
                self.results.append_all_results(current_sim_results)
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # Implement the _on_simulate_finish method in a subclass if you
            # need to run code at the end of the simulate method.
            self._on_simulate_finish()

            # xxxxxxx Save the number of runned iterations xxxxxxxxxxxxxxxx
            self.results.runned_reps = self._runned_reps
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Update the elapsed time xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            self.__toc = time()
            self._elapsed_time = self.__toc - self.__tic

            # Also save the elapsed time in the SimulationResults object
            self.results.elapsed_time = self._elapsed_time
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Save the results if results_filename is not None xxxxxx
            if self._results_base_filename is not None:
                self.results.save_to_file(self._results_base_filename)
                # Delete the partial results (this will only delete the
                # partial results if self.delete_partial_results_bool is
                # True)
                self.__delete_partial_results_maybe()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # The unittests for this method only run if an ipython cluster is
    # started with a profile called "tests".
    def simulate_in_parallel(self,
                             view: ParallelView,
                             wait: bool = True) -> None:  # pragma: no cover
        """
        Same as the simulate method, but the different parameters
        configurations are simulated in parallel.

        Parameters
        ----------
        view : LoadBalancedView | DirectView
            A ´view´ of the IPython engines.
            The parallel processing will happen by calling the 'map' method
            of the provided view to simulate in parallel the different
            configurations of transmission parameters.
        wait : bool
            If True then the self.wait_parallel_simulation method will be
            automatically called at the end of simulate_in_parallel. If
            False, the YOU NEED to manually call
            self.wait_parallel_simulation at some point after calling
            simulate_in_parallel.

        Notes
        -----
        There is a limitation regarding the partial simulation results. The
        partial results files will be saved in the folder where the IPython
        engines are running, since the "saving part" is performed in an
        IPython engine. However, the deletion of the files is not performed
        in by the IPython engines, but by the main python
        program. Therefore, unless the Ipython engines are running in the
        same folder where the main python program will be run the partial
        result files won't be automatically deleted after the simulation is
        finished.
        """
        # xxxxxxxxxxxxxxx Some initialization xxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        self.__tic = time()
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxx Store rep_max in the results object xxxxxxxxxxxxxxxxxxxxxxx
        self.results.rep_max = self.rep_max  # pylint: disable=W0201
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Store the Simulation parameters in the SimulationResults object.
        # With this, the simulation parameters will be available for
        # someone that has the SimulationResults object (loaded from a
        # file, for instance).
        self.results.set_parameters(self.params)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Implement the _on_simulate_start method in a subclass if you need
        # to run code at the start of the simulate method.
        self._on_simulate_start()
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Get the number of variations of the transmit parameters
        num_variations = self.params.get_num_unpacked_variations()

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # xxxxx Start of the code unique to the parallel version xxxxxxxxx
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxx Progressbar for the parallel simulation xxxxxxxxxxxxxx
        if self.update_progress_function_style is not None:  # pragma: no cover
            # Create the proxy progressbars
            proxybar_data_list = []
            for _ in range(num_variations):
                proxybar_data_list.append(
                    self._get_parallel_update_progress_function())
        else:  # self.update_progress_function_style is None
            # Create the dummy update progress functions
            proxybar_data_list = [None] * num_variations
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxx Perform the actual simulation in asynchronously parallel xxxx
        # NOTE: If this fails because of some pickling error, make sure the
        # class of 'self' (that is, the subclass of SimulationRunner that
        # you are trying to run) is pickle-able.
        self._async_results = view.map(
            # simulate_for_current_params,
            SimulationRunner._simulate_for_current_params_parallel,
            # We need to pass the SimulationRunner
            # object to the IPython engine ...
            [self] * num_variations,
            # ... and we also need to pass the
            # simulation parameters for each engine
            self.params.get_unpacked_params_list(),
            proxybar_data_list,
            block=False)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        if self._pbar is not None:  # pragma: no cover
            self._pbar.start_updater()

        if wait is True:
            self.wait_parallel_simulation()

    def wait_parallel_simulation(self) -> None:  # pragma: no cover
        """
        Wait for the parallel simulation to finish and then update the
        self.results variable (as well as other internal variables).
        """
        # Note that at the end of this method we set self._async_results to
        # None. Therefore, if wait_parallel_simulation is called multiple
        # times nothing will happen.
        if self._async_results is not None:
            # Wait for the tasks (running in the IPython engines) to finish
            self._async_results.wait()

            results = self._async_results.get()
            for reps, r, filename in results:
                self._runned_reps.append(reps)
                self.results.append_all_results(r)
                self._results_base_filename_unpack_list.append(filename)

            # Implement the _on_simulate_finish method in a subclass if you
            # need to run code at the end of the simulate method.
            self._on_simulate_finish()

            # xxxxxxx Save the number of runned iterations xxxxxxxxxxxxxxxx
            self.results.runned_reps = self.runned_reps
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Update the elapsed time xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            self.__toc = time()
            self._elapsed_time = self.__toc - self.__tic

            # Also save the elapsed time in the SimulationResults object
            self.results.elapsed_time = self.elapsed_time
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # xxxxx Save the results if results_base_filename is not None x
            if self._results_base_filename is not None:
                self.results.save_to_file(self.results_base_filename)
                # Delete the partial results (this will only delete the
                # partial results if self.delete_partial_results_bool is
                #  True)
                self.__delete_partial_results_maybe()
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # Stop the progressbar updating progress
            if self.update_progress_function_style is not None:
                # pragma: no cover
                self._pbar.stop_updater()

            # Erase the self._async_results object, since we already got
            #  all information we needed from it
            self._async_results = None

    # noinspection PyMethodMayBeStatic
    def _on_simulate_start(self) -> None:
        """This method is called only once, in the beginning of the the
        simulate method.

        """

    # noinspection PyMethodMayBeStatic
    def _on_simulate_finish(self) -> None:
        """This method is called only once at the end of the simulate method.

        """

    def _on_simulate_current_params_start(
            self, current_params: SimulationParameters) -> None:
        """
        This method is called once for each simulation parameters
        combination before any iteration of _run_simulation is performed
        (for that combination of simulation parameters).

        Parameters
        ----------
        current_params : SimulationParameters
            The current combination of simulation parameters.

        Notes
        -----
        Because this method will run inside the _run_simulation method,
        which is called in a different process when the simulation is
        performed in parallel, it should only modify member variables that
        are only used inside _run_simulation. If any other variable is
        modified these changes WILL NOT be carried back to the original
        process.
        """

    # noinspection PyMethodMayBeStatic
    def _on_simulate_current_params_finish(
            self, current_params: SimulationParameters,
            current_params_sim_results: SimulationResults) -> None:
        """This method is called once for each simulation parameters
        combination after all iterations of _run_simulation are performed
        (for that combination of simulation parameters).

        Parameters
        ----------
        current_params : SimulationParameters
            The current combination of simulation parameters.
        current_params_sim_results : SimulationResults
            SimulationResults object with the results for the finished
            simulation with the parameters in current_params.

        Notes
        -----
        Because this method will run inside the _run_simulation method,
        which is called in a different process when the simulation is
        performed in parallel, it should only modify member variables that
        are only used inside _run_simulation.
        """


# xxxxxxxxxx SimulationRunner - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx



#!/usr/bin/env python
"""Module implementing helper functions for simulators."""

import sys
from typing import Any, List, Optional, Union

from .progressbar import ProgressbarZMQServer
from .runner import SimulationRunner

try:
    # noinspection PyUnresolvedReferences
    from ipyparallel import Client, LoadBalancedView, DirectView
except ImportError:  # pragma: no cover
    Client = Any
    LoadBalancedView = Any
    DirectView = Any


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Module functions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def simulate_do_what_i_mean(
        runner_or_list_of_runners: Union[SimulationRunner,
                                         List[SimulationRunner]],
        folder: Optional[str] = None) -> None:  # pragma: no cover
    """
    This will either call the simulate method or the simulate_in_parallel
    method as appropriated.

    If the 'parameters variation index' was specified in the command
    line, then the 'simulate' method will be called with that index. If
    not, then the simulate method will be called without any index or,
    if there is an ipython cluster running, the simulate_in_parallel
    method will be called.

    Parameters
    ----------
    runner_or_list_of_runners : SimulationRunner | list[SimulationRunner]
        The SimulationRunner object for which either the 'simulate' or the
        'simulate_in_parallel' method will be called. If this is a list,
        then we just call this method individually for each member of the
        list.
    folder : str
        Folder to be added to the python path. This should be the main
        pyphysim folder
    """
    if isinstance(runner_or_list_of_runners, list):
        list_of_runners = runner_or_list_of_runners
        _simulate_do_what_i_mean_multiple_runners(list_of_runners, folder)
    else:
        # xxxxxxxxxx We only have one SimulationRunner obj xxxxxxxxxxxxxxxx
        runner = runner_or_list_of_runners
        _simulate_do_what_i_mean_single_runner(runner, folder, block=True)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


def _add_folder_to_ipython_engines_path(
        client: Client, folder: str) -> None:  # pragma: no cover
    """
    Add a folder to sys.path of each ipython engine.

    The list of engines is get as a direct view from 'client'.

    This will also add the folder to the local python path.

    Parameters
    ----------
    client : Client
        The client from which we will get a direct view to access the
        engines.
    folder : str
        The folder to be added to the python path at each engine.
    """
    # Add the folder to the python path of the main application
    sys.path.append(folder)

    # We create a direct view to run coe in all engines
    dview = client.direct_view()
    # Reset the engines so that we don't have variables there from last
    # computations
    dview.execute('%reset')
    dview.execute('import sys')
    # Add the folder to the python path of each engine. We use
    # block=True to ensure that all engines have modified their
    # path to include the folder with the simulator before we
    # continue.
    dview.execute('sys.path.append("{0}")'.format(folder), block=True)


def _simulate_do_what_i_mean_single_runner(
        runner: SimulationRunner,
        folder: Optional[str] = None,
        block: bool = True) -> None:  # pragma: no cover
    """
    This will either call the `simulate` method or the
    `simulate_in_parallel` method as appropriated.

    If the 'parameters variation index' was specified in the command line,
    then the `simulate` method will be called with that index. If not, then
    the `simulate` method will be called without any index or, if there is
    an ipython cluster running, the `simulate_in_parallel` method will be
    called.

    Parameters
    ----------
    runner : SimulationRunner
        The SimulationRunner object for which either the 'simulate' or the
        'simulate_in_parallel' method will be called.
    folder : str, optional
        Folder to be added to the python path. This should be the main
        pyphysim folder
    block : bool, optional
        Passed to the simulate_in_parallel method when the simulation is
        performed in parallel. If this is false, you need to call the
        method 'wait_parallel_simulation' of the runner object at some
        point.
    """
    if runner.command_line_args.index is not None:
        # Perform the simulation (serially) for the desired index
        msg = "Simulation will be run for the parameters variation: {0}"
        print(msg.format(runner.command_line_args.index))
        runner.simulate(runner.command_line_args.index)

    else:
        run_in_parallel = True
        try:
            # If we can get an IPython view that means that the IPython
            # engines are running. In that case we will perform the
            # simulation in parallel
            from ipyparallel import Client
            # cl = Client(profile="ssh")
            # noinspection PyTypeChecker
            cl = Client(profile="default", timeout=2.0)

            if folder is not None:
                _add_folder_to_ipython_engines_path(cl, folder)

            # For the actual simulation we are better using a
            # load_balanced_view
            lview = cl.load_balanced_view()
        except (IOError, ImportError):
            # If we can't get an IPython view then we will perform the
            # simulation serially
            run_in_parallel = False
            lview = None

        if run_in_parallel is True:
            print("Simulation will be run in Parallel")

            # Remove the " - SNR: {SNR}" string in the progressbar message,
            # since when the simulation is performed in parallel we get a
            # single progressbar for all the simulation.
            runner.progressbar_message = 'Elapsed Time: {{elapsed_time}}'
            runner.simulate_in_parallel(lview, wait=block)
        else:
            print("Simulation will be run serially")
            runner.simulate()


def _simulate_do_what_i_mean_multiple_runners(
        list_of_runners: List[SimulationRunner],
        folder: Optional[str] = None) -> None:  # pragma: no cover
    """
    This will either call the `simulate` method or the
    `simulate_in_parallel` method as appropriated.

    If the 'parameters variation index' was specified in the command line,
    then the `simulate` method will be called with that index. If not, then
    the `simulate` method will be called without any index or, if there is
    an ipython cluster running, the `simulate_in_parallel` method will be
    called.

    Parameters
    ----------
    list_of_runners : list[SimulationRunner]
        The `_simulate_do_what_i_mean_single_runner` will be called for
        each object in the list.
    folder : str
        Folder to be added to the python path. This should be the main
        pyphysim folder.
    """
    # If we have a list of SimulationRunner objects, we want two
    # things. First, we want to use the same progressbar for all of
    # them. Second, we want to use 'block=False' for all of them and
    # only later call the wait_parallel_simulation method for each
    # runner.

    # First we will check the progress_output_type variable of the first
    # runner. If it is 'screen' we will create a progressbar that prints
    # the progress to the terminal, otherwise we will create a progressbar
    # the prints the progress to a file.
    if list_of_runners[0].progress_output_type == 'screen':
        # Progress will be printed to the screen
        filename = None
    else:
        # Progress will be printed to this file
        filename = 'multiple_runners_progress.txt'

    # xxxxxxxxxx Create the shared progressbar object xxxxxxxxxxxxxxxxx
    pbar = ProgressbarZMQServer(progresschar='*',
                                message="Elapsed Time: {elapsed_time}",
                                sleep_time=1,
                                filename=filename)
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # xxxxxxxxxx Start the simulation for each runner xxxxxxxxxxxxxxxxx
    add_folder_to_path = True
    for runner in list_of_runners:
        runner._pbar = pbar  # pylint: disable= W0212
        if add_folder_to_path is True:
            _simulate_do_what_i_mean_single_runner(runner, folder, block=False)
            add_folder_to_path = False
        else:
            _simulate_do_what_i_mean_single_runner(runner, None, block=False)
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # xxxxx Wait for the simulation of each runner to finish xxxxxxxxxx
    for runner in list_of_runners:
        runner.wait_parallel_simulation()
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxx Module Functions - END xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
